"""
.. module:: hkSeqYeck
        :synopsis: Sequential H-K stacking method for receiver function analysis after Yeck et al., 2013
        Yeck, W. L., Sheehan, A. F., & Schulte-Pelkum, V. (2013, 6). Sequential h-k stacking to obtain accurate 
        crustal thicknesses beneath sedimentary basins. Bulletin of the Seismological Society of America, 103 , 
        2142-2150. doi: 10.1785/0120120290
.. moduleauthor:: Stephen Akinremi <s.akinremi@utwente.nl> Islam Fadel <i.e.a.m.fadel@utwente.nl> (October 2023)
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.transforms as mtransforms
from matplotlib.legend_handler import HandlerLine2D as HL
import matplotlib.gridspec as gridspec
from obspy.geodetics.base import kilometer2degrees
import seaborn
seaborn.set_style("darkgrid", {"axes.facecolor": ".85"})
#----------------------------------------------------------
def getamp(rfdata, tarray, t):
    """
    Get the amplitude of the receiver function at a specific time

    :param rfdata: receiver function data
    :type rfdata: numpy.ndarray
    :param tarray: time array
    :type tarray: numpy.ndarray
    :param t: time to get the amplitude
    :type t: float
    :return: amplitude of the receiver function at time t
    """
    amp = rfdata[(np.abs(tarray - t).argmin())]
    return amp
#----------------------------------------------------------
def hkSeq(rfstreamSed, rfstreamMoho, HSed=np.linspace(0,10,201), KSed=np.linspace(1.65,2.25,201), VpSed=3.0,
                 w1Sed=0.6, w2Sed=0.3, w3Sed=0.1, HMoho=np.linspace(20,60,201), KMoho=np.linspace(1.65,1.95,121), VpMoho=6.9, 
                 w1Moho=0.6, w2Moho=0.3, w3Moho=0.1, g = [75.,10., 15., 2.5], stack=None):
    """
    Sequential H-K stacking method for receiver function analysis after Yeck et al., 2013
    
    :param rfstreamSed: Stream object of receiver function for sediment layer (high frequency)
    :type rfstreamSed: obspy.core.stream.Stream
    :param rfstreamMoho: Stream object of receiver function for Moho layer (lower frequency)
    :type rfstreamMoho: obspy.core.stream.Stream
    :param HSed: Depth range for sediment layer
    :type HSed: numpy.ndarray
    :param KSed: Vp/Vs range for sediment layer
    :type KSed: numpy.ndarray
    :param VpSed: Vp for sediment layer
    :type VpSed: float
    :param w1Sed: Weight for Ps arrival
    :param w2Sed: Weight for PpPs arrival
    :param w3Sed: Weight for PsPs arrival
    :type w1Sed,w2Sed,w3Sed: float
    :param HMoho: Depth range for Moho layer
    :type HMoho: numpy.ndarray
    :param KMoho: Vp/Vs range for Moho layer
    :type KMoho: numpy.ndarray
    :param VpMoho: Vp for Moho layer
    :type VpMoho: float
    :param w1Moho: Weight for Ps arrival
    :param w2Moho: Weight for PpPs arrival
    :param w3Moho: Weight for PpSs+PsPs arrival
    :type w1Moho,w2Moho,w3Moho: float
    :param g: Gain for plotting
    :type g: list
    :param rmneg: Remove negative values in the stacking results
    :type rmneg: bool
    :param stack: Stack the receiver functions traces into one before hk stacking
    :type stack: bool
    :return: SequentialHKResult: Dictionary of results from the Sequential H-K stacking method
    """
    if g is None: g = [75.,10., 15., 2.5]
    if VpSed is None: VpSed = 3.0
    if w1Sed is None: w1Sed = 0.6
    if w2Sed is None: w2Sed = 0.3
    if w3Sed is None: w3Sed = 0.1  
    if stack is None: stack = False 
    StackedRFSed=rfstreamSed.stack()
    RFStackedSed=StackedRFSed[0].data
    StackedRFMoho=rfstreamMoho.stack()
    RFStackedMoho=StackedRFMoho[0].data
    staname = rfstreamSed[0].stats.station
    comp =  rfstreamSed[0].stats.channel
    rfdata = rfstreamSed[0].data
    slowness=[]
    for tr in rfstreamMoho:
        slow=tr.stats['slowness']
        slowness.append(slow)
    AvgSlow = np.mean(slowness)
    delta = rfstreamSed[0].stats.delta
    l = len(rfdata)
    t = np.arange(0, l)
    t = (delta *  t) - 10 
    #----------------------------------------------------------
    # Stack for Sediment Layer
    #----------------------------------------------------------
    stkSed = np.zeros((len(KSed)*len(HSed),3))
    z = 0 
    for i in range(len(KSed)):
        Ktemp = KSed[i]
        for j in range(len(HSed)):
            Htemp = HSed[j]
            s = 0.0
            for tr in rfstreamSed:
                trdata = tr.data
                b = tr.stats.starttime - tr.stats.onset
                delta= tr.stats.delta
                rp = kilometer2degrees(1) * tr.stats['slowness']
                if stack == True:
                    trdata = StackedRFSed[0].data
                    rp = kilometer2degrees(1) * AvgSlow
                #--------------------------------------
                term1= ((Ktemp/VpSed)**2 - rp**2)**0.5
                term2= ((1/VpSed)**2 - rp**2)**0.5
                tpsSed = Htemp * (term1 - term2)
                tpppsSed = Htemp * (term1 + term2)
                tpspsSed = Htemp * 2 * (term1) 
                #--------------------------------------
                ampps = getamp(trdata, t, tpsSed)
                ampppps = getamp(trdata,t, tpppsSed)
                amppsps = getamp(trdata, t, tpspsSed)
                #--------------------------------------
                stemp = (w1Sed * ampps) + (w2Sed * ampppps) - (w3Sed * amppsps)
                sSed = stemp + s
            stkSed[z,:] = [Htemp, Ktemp, sSed]
            z = z + 1
    bmodSed = stkSed[np.argmax(stkSed[:,2]),:]
    #----------------------------------------------------------
    # Plot Sediment Stacking results
    plt.tricontourf(stkSed[:,0], stkSed[:,1], stkSed[:,2],60, cmap='jet')
    cb=plt.colorbar(format='%.3f', orientation="vertical")
    p1,=plt.plot(bmodSed[0],bmodSed[1], 'k+', mew=5, ms=15,\
        label='Best Model %s km %s Vp/Vs'%(bmodSed[0], bmodSed[1]))
    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,\
        ncol=2, mode="expand", borderaxespad=0, \
        handler_map={p1:HL(numpoints=1)}) 
    plt.ylabel('Vp/Vs')
    plt.xlabel('Depth km')
    plt.xlim(min(HSed),max(HSed))
    plt.ylim(min(KSed),max(KSed))
    # plt.savefig(savepath , format=format, dpi=250) 
    plt.show()
    plt.close("all")    
    #----------------------------------------------------------
    # Add Ps, PpPs, PsPs + PbSs timing to the data header
    #----------------------------------------------------------    
    H1 = bmodSed[0]
    K1 = bmodSed[1]
    SSed = bmodSed[2]
    print("Best Sediment depth: ", H1, "Best Sediment Vp/Vs:", K1, "Max stack: ", SSed)
    #----------------------------------------------------------
    # Stack for Moho Layer
    #----------------------------------------------------------
    stkMoho = np.zeros((len(KMoho)*len(HMoho),3))
    z = 0 
    for i in range(len(KMoho)):
        Ktemp = KMoho[i]
        for j in range(len(HMoho)):
            Htemp = HMoho[j]
            s = 0.0
            for tr in rfstreamMoho:
                trdata = tr.data
                b = rfstreamMoho[0].stats.starttime - rfstreamMoho[0].stats.onset
                delta= tr.stats.delta
                rp = kilometer2degrees(1) * tr.stats['slowness']
                if stack == True:
                    trdata = StackedRFMoho[0].data
                    rp=kilometer2degrees(1) * AvgSlow 
                #--------------------------------------
                term1 = ((K1/VpSed)**2-(rp)**2)**0.5
                term2 = ((1/VpSed)**2-(rp)**2)**0.5
                term3 = ((Ktemp/VpMoho)**2-(rp)**2)**0.5
                term4 = ((1/VpMoho)**2-(rp)**2)**0.5
                #--------------------------------------
                tpsMoho = (H1 * (term1 - term2)) + (Htemp * (term3 - term4))
                tpppsMoho = (H1 * (term1 + term2)) + (Htemp * (term3 + term4))
                tpspsMoho = (2 * H1 * term1) + (2 * Htemp * term3)
                #--------------------------------------
                ampps = getamp(trdata, t, tpsMoho) 
                ampppps = getamp(trdata, t, tpppsMoho)  
                amppsps = getamp(trdata, t, tpspsMoho)
                stemp = (w1Moho * ampps) + (w2Moho * ampppps) - (w3Moho * amppsps)
                sMoho = stemp + s
                #--------------------------------------
            stkMoho[z,:] = [Htemp, Ktemp, sMoho]
            z = z + 1
    bmodMoho = stkMoho[np.argmax(stkMoho[:,2]),:]
    #----------------------------------------------------------
    # plot Moho Stacking results
    for i in range(len(stkMoho)):
        if stkMoho[i,2] <= 0:
            stkMoho[i,2] = 0
    plt.tricontourf(stkMoho[:,0], stkMoho[:,1], stkMoho[:,2],60, cmap='jet')
    cb=plt.colorbar(format='%.3f', orientation="vertical")
    p1,=plt.plot(bmodMoho[0],bmodMoho[1], 'k+', mew=5, ms=15,\
        label='Best Model %s km %s Vp/Vs'%(bmodMoho[0], bmodMoho[1]))
    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,\
        ncol=2, mode="expand", borderaxespad=0, \
        handler_map={p1:HL(numpoints=1)}) 
    plt.ylabel('Vp/Vs')
    plt.xlabel('Depth km')
    plt.xlim(min(HMoho),max(HMoho))
    plt.ylim(min(KMoho),max(KMoho))
    # plt.savefig(savepath , format=format, dpi=250) 
    plt.show()
    plt.close("all")    
    H2 = bmodMoho[0]
    K2 = bmodMoho[1] 
    print("Best Moho depth: ", H2, "Best Moho Vp/Vs:", K2, "Max stack: ", SSed)   
    SequentialHKResult = {'RFSed': rfstreamSed, 'RFSedStack': RFStackedSed, 'RFMoho': rfstreamMoho, 'RFMohoStack': RFStackedMoho, 'time':t, \
                          'bestmodelSed':bmodSed, 'stackvalueSed':stkSed, 'bestmodelMoho':bmodMoho, 'stackvalueMoho':stkMoho,
                         'stack':stack, 'comp':comp, 'AvgSlow':AvgSlow, 'VpSed':VpSed, 'VpMoho':VpMoho, 'HSed':HSed, 'KSed':KSed, 'HMoho':HMoho, 'KMoho':KMoho, 'staname':staname}
    return SequentialHKResult  
#----------------------------------------------------------
def plotSeqhk(SequentialHKResult, savepath, g = [75.,10., 15., 2.5], rmneg = True, format = 'jpg'): 
    """
    Plot the results from the Sequential H-K stacking method

    :param SequentialHKResult: Dictionary of results from the Sequential H-K stacking method (function: hkSequential)
    :type SequentialHKResult: dict
    :param g: Gain for plotting
    :type g: list
    :param rmneg: Remove negative values in the stacking results
    :type rmneg: bool
    :param format: Format of the figure
    :type format: str
    :param savepath: Path to save the figure
    :type savepath: str
    """
    if rmneg is None: rmneg = True
    if format is None: format = 'pdf'
    if g is None: g = [75.,10., 15., 2.5]
    staname=SequentialHKResult['staname']
    comp=SequentialHKResult['comp']
    stack=SequentialHKResult['stack']
    VpSed=SequentialHKResult['VpSed']
    HSed=SequentialHKResult['HSed']
    KSed=SequentialHKResult['KSed']
    bmodSed = SequentialHKResult['bestmodelSed']
    H1 = bmodSed[0]
    K1 = bmodSed[1]
    stkSed=SequentialHKResult['stackvalueSed']
    if SequentialHKResult['stack'] == True:
        rfstream=SequentialHKResult['RFSedStack']
        rp = kilometer2degrees(1) * SequentialHKResult['AvgSlow']
        t=SequentialHKResult['time']
        term1= ((K1/VpSed)**2 - rp**2)**0.5
        term2= ((1/VpSed)**2 - rp**2)**0.5
        tpsSed = H1 * (term1 - term2)
        tpppsSed = H1 * (term1 + term2)
        tpspsSed = H1 * 2 * (term1)
    else:
        rfstreamSed=SequentialHKResult['RFSed']
        for tr in rfstreamSed:
            rp = kilometer2degrees(1) * tr.stats['slowness']
            term1= ((K1/VpSed)**2 - rp**2)**0.5
            term2= ((1/VpSed)**2 - rp**2)**0.5
            tpsSed = H1 * (term1 - term2)
            tpppsSed = H1 * (term1 + term2)
            tpspsSed = H1 * 2 * (term1)
            tr.stats['tpsSed'] = tpsSed
            tr.stats['tpppsSed'] = tpppsSed
            tr.stats['tpspsSed'] = tpspsSed
    
    #----------------------------------------------------------
    # plot H-K results
    #----------------------------------------------------------
    plt.figure(figsize=(18, 6)) 
    gs = gridspec.GridSpec(1, 3, width_ratios=[6,4,4])
    #----------------------------------------------------------
    # plot misfit
    #----------------------------------------------------------
    #-----Remove the stack values that is lower than zero------
    if rmneg == True:
        for i in range(len(stkSed)):
            if stkSed[i,2] <= 0:
                stkSed[i,2] = 0
    #----------------------------------------------------------
    ax1 = plt.subplot(gs[0, 0])
    plt.tricontourf(stkSed[:,0], stkSed[:,1], stkSed[:,2],50, cmap ="jet")
    #----------------------------------------------------------
    cb=plt.colorbar(format='%.3f', orientation="vertical")
    p1,=ax1.plot(bmodSed[0],bmodSed[1], 'k+', mew=2, ms=10, \
    label='%s Best Model %.2f km %.2f Vp/Vs'%(staname,bmodSed[0],bmodSed[1]+0.001))
    props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)

    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,ncol=2, mode="expand", borderaxespad=0, handler_map={p1:HL(numpoints=1)})
    ax1.set_ylabel('Vp/Vs')
    ax1.set_xlabel('Depth km')
    ax1.set_xlim(min(HSed), max(HSed))
    ax1.set_ylim(min(KSed), max(KSed))
    #----------------------------------------------------------
    # Plot the receiver function witht the estimated times
    #----------------------------------------------------------
    ax2 = plt.subplot(gs[0, 1]) 
    
    if stack == True:
        baz=180
        rfdata = ((SequentialHKResult['RFSedStack']) * g[0]) + baz
        major_ticks_x = np.arange(-10, 41, 1)                                              
        minor_ticks_x = np.arange(-10, 41, 1) 
        major_ticks_y = np.arange(0, 361, 45)                                              
        minor_ticks_y = np.arange(-30, 390, 5)
        ax2.set_xticks(major_ticks_x)                                                       
        ax2.set_xticks(minor_ticks_x, minor=True)                                           
        ax2.set_yticks(major_ticks_y)                                                       
        ax2.set_yticks(minor_ticks_y, minor=True)                                          
        ax2.set_xlim(-1, 10)
        ax2.set_ylim(-29, 390)
        ax2.set_xlabel("Time (sec)")
        ax2.set_ylabel("Back-Azimuth (deg)")  
        ax2.plot(t, rfdata, "k-", lw=0.5)
        ax2.fill_between(t, baz, rfdata, where=rfdata > baz, facecolor='red', alpha = 0.25)
        ax2.fill_between(t, baz, rfdata, where=rfdata <= baz, facecolor='blue', alpha = 0.25)
        mtransforms.blended_transform_factory(ax2.transData, ax2.transAxes)
        ampgain = g[1]
        ax2.plot([tpsSed, tpsSed], [baz+ampgain, baz-ampgain], 'g-', lw=1.5, label="Moho")
        ax2.plot([tpppsSed, tpppsSed], [baz+ampgain, baz-ampgain], 'g-', lw=1.5)
        ax2.plot([tpspsSed, tpspsSed], [baz+ampgain, baz-ampgain], 'g-', lw=1.5)
        filename = "%s/%s_%s_%s.%s"%(savepath, staname, comp, "Sediment", format)
        plt.savefig(filename , format=format, transparent=False,\
            dpi=250, bbox_inches = 'tight', pad_inches=0.1)
        plt.show()
        plt.close('all')
    else:
        rfstreamSed = SequentialHKResult['RFSed']
        gcarcarr = np.zeros(len(rfstreamSed))
        for rf in rfstreamSed:        
            delta = rf.stats.delta
            baz = rf.stats.back_azimuth
            b = rf.stats.starttime - rf.stats.onset
            tpsSed = rf.stats.tpsSed
            tpppsSed = rf.stats.tpppsSed
            tpspsSed = rf.stats.tpspsSed
            rfdata = ((rf.data) * g[0]) + baz
            t = (np.arange(0, len(rfdata), 1) * delta) + b
            major_ticks_x = np.arange(-10, 41, 1)                                              
            minor_ticks_x = np.arange(-10, 41, 1) 
            major_ticks_y = np.arange(0, 361, 45)                                              
            minor_ticks_y = np.arange(-30, 390, 5)
            ax2.set_xticks(major_ticks_x)                                                       
            ax2.set_xticks(minor_ticks_x, minor=True)                                           
            ax2.set_yticks(major_ticks_y)                                                       
            ax2.set_yticks(minor_ticks_y, minor=True)                                          
            ax2.set_xlim(-1, 10)
            ax2.set_ylim(-29, 390)
            ax2.plot(t, rfdata, "k-", lw=0.5)
            ax2.fill_between(t, baz, rfdata, where=rfdata > baz, facecolor='red', alpha = 0.25)
            ax2.fill_between(t, baz, rfdata, where=rfdata <= baz, facecolor='blue', alpha = 0.25)
            mtransforms.blended_transform_factory(ax2.transData, ax2.transAxes)
            #----------------------------------------------------------
            # Plot H-K sequential formula times
            #----------------------------------------------------------
            ampgain = g[1]
            ax2.plot([tpsSed, tpsSed], [baz+ampgain, baz-ampgain], 'g-', lw=1.5, label="Moho")
            ax2.plot([tpppsSed, tpppsSed], [baz+ampgain, baz-ampgain], 'g-', lw=1.5)
            ax2.plot([tpspsSed, tpspsSed], [baz+ampgain, baz-ampgain], 'g-', lw=1.5)
        #----------------------------------------------------------
        # Highlight specifiec time range on RF plot
        #----------------------------------------------------------
        ax2.axvspan(tpsSed-0.1, tpsSed+0.1, alpha=0.5, color='lightgreen')
        ax2.axvspan(tpppsSed-0.1, tpppsSed+0.1, alpha=0.5, color='lightgreen')
        ax2.axvspan(tpspsSed-0.1, tpspsSed+0.1, alpha=0.5, color='lightgreen')
        props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
        ax2.set_xlabel("Time (sec)")
        ax2.set_ylabel("Back-Azimuth (deg)")    
        #----------------------------------------------------------
        # Plot the receiver function witht the estimated times
        #----------------------------------------------------------
        ax3 = plt.subplot(gs[0, 2])
        gcarcarr = np.zeros(len(rfstreamSed))   
        for rf, j in zip(rfstreamSed, np.arange(len(rfstreamSed))):        
            delta = rf.stats.delta
            gcarc = rf.stats.distance
            gcarcarr[j] = gcarc
            b = rf.stats.starttime - rf.stats.onset
            tpsSed = rf.stats.tpsSed
            tpppsSed = rf.stats.tpppsSed
            tpspsSed = rf.stats.tpspsSed
            rfdata = ((rf.data) * g[2]) + gcarc
            t = (np.arange(0, len(rfdata), 1) * delta) + b    
            major_ticks_x = np.arange(-10, 41, 1)                                              
            minor_ticks_x = np.arange(-10, 41, 1) 
            major_ticks_y = np.arange(30, 110, 10)                                              
            minor_ticks_y = np.arange(30, 111, 5)
            ax3.set_xticks(major_ticks_x)                                                       
            ax3.set_xticks(minor_ticks_x, minor=True)                                           
            ax3.set_yticks(major_ticks_y)                                                       
            ax3.set_yticks(minor_ticks_y, minor=True)                                          
            #----------------------------------------------------------
            # Plot grid on the figure
            #----------------------------------------------------------                                                              
            ax3.set_xlim(-1, 10)
            ax3.set_ylim(25,110)
            ax3.plot(t, rfdata, "k-", lw=0.5)
            ax3.fill_between(t, gcarc, rfdata, where=rfdata > gcarc, facecolor='red', alpha = 0.25)
            ax3.fill_between(t, gcarc, rfdata, where=rfdata <= gcarc, facecolor='blue', alpha = 0.25)
            mtransforms.blended_transform_factory(ax3.transData, ax3.transAxes)
            #----------------------------------------------------------
            # Plot H-K sequential formula times
            #----------------------------------------------------------
            ampgain = g[3]
            ax3.plot([tpsSed, tpsSed], [gcarc+ampgain, gcarc-ampgain], 'g-', lw=1.5, label="Moho")
            ax3.plot([tpppsSed, tpppsSed], [gcarc+ampgain, gcarc-ampgain], 'g-', lw=1.5)
            ax3.plot([tpspsSed, tpspsSed], [gcarc+ampgain, gcarc-ampgain], 'g-', lw=1.5)
        #----------------------------------------------------------
        # Highlight specifiec time range on RF plot
        #----------------------------------------------------------
        ax3.axvspan(tpsSed-0.1, tpsSed+0.1, alpha=0.5, color='lightgreen')
        ax3.axvspan(tpppsSed-0.1, tpppsSed+0.1, alpha=0.5, color='lightgreen')
        ax3.axvspan(tpspsSed-0.1, tpspsSed+0.1, alpha=0.5, color='lightgreen')    

        props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)

        ax3.set_xlabel("Time (sec)")
        ax3.set_ylabel("Distance (deg)")
        
        filename = "%s/%s_%s_%s.%s"%(savepath, staname, comp, "Sediment", format)
        plt.savefig(filename , format=format, transparent=False,\
            dpi=250, bbox_inches = 'tight', pad_inches=0.1)
        plt.show()
        plt.close('all')
    #----------------------------------------------------------
    # Plot Moho Stack
    #----------------------------------------------------------
    VpMoho=SequentialHKResult['VpMoho']
    HMoho=SequentialHKResult['HMoho']
    KMoho=SequentialHKResult['KMoho']
    bmodMoho = SequentialHKResult['bestmodelMoho']  
    H2 = bmodMoho[0]
    K2 = bmodMoho[1]
    stkMoho=SequentialHKResult['stackvalueMoho']

    if SequentialHKResult['stack'] == True:
        rfstream=SequentialHKResult['RFMohoStack']
        rp = kilometer2degrees(1) * SequentialHKResult['AvgSlow']
        t=SequentialHKResult['time']
        term1 = ((K1/VpSed)**2-(rp)**2)**0.5    
        term2 = ((1/VpSed)**2-(rp)**2)**0.5
        term3 = ((K2/VpMoho)**2-(rp)**2)**0.5    
        term4 = ((1/VpMoho)**2-(rp)**2)**0.5
        #--------------------------------------
        tpsMoho = (H1 * (term1 - term2)) + (H2 * (term3 - term4))
        tpppsMoho = (H1 * (term1 + term2)) + (H2 * (term3 + term4))
        tpspsMoho = (2 * H1 * term1) + (2 * H2 * term3)
    else:
        rfstreamMoho=SequentialHKResult['RFMoho']
        for tr in rfstreamMoho:
            rp = kilometer2degrees(1) * tr.stats['slowness']
            #--------------------------------------
            term1 = ((K1/VpSed)**2-(rp)**2)**0.5    
            term2 = ((1/VpSed)**2-(rp)**2)**0.5
            term3 = ((K2/VpMoho)**2-(rp)**2)**0.5    
            term4 = ((1/VpMoho)**2-(rp)**2)**0.5
            #--------------------------------------
            tpsMoho = (H1 * (term1 - term2)) + (H2 * (term3 - term4))
            tpppsMoho = (H1 * (term1 + term2)) + (H2 * (term3 + term4))
            tpspsMoho = (2 * H1 * term1) + (2 * H2 * term3)
            #--------------------------------------      
            tr.stats['tpsMoho'] = tpsMoho
            tr.stats['tpppsMoho'] = tpppsMoho
            tr.stats['tpspsMoho'] = tpspsMoho
    plt.figure(figsize=(18, 6)) 
    gs = gridspec.GridSpec(1, 3, width_ratios=[6,4,4])
    #-----Remove the stack values that is lower than zero------
    if rmneg == True:
        for i in range(len(stkMoho)):
            if stkMoho[i,2] <= 0:
                stkMoho[i,2] = 0
        #----------------------------------------------------------
    ax1 = plt.subplot(gs[0, 0])
    plt.tricontourf(stkMoho[:,0], stkMoho[:,1], stkMoho[:,2],50, cmap ="jet")
    cb=plt.colorbar(format='%.3f', orientation="vertical")
    cb.ax.set_xlabel('S', rotation=0)
    p1,=ax1.plot(bmodMoho[0],bmodMoho[1], 'k+', mew=2, ms=10, \
    label='%s Best Model %.2f km %.2f Vp/Vs'%(staname,bmodMoho[0],bmodMoho[1]+0.001))
    props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,\
        ncol=2, mode="expand", borderaxespad=0, \
        handler_map={p1:HL(numpoints=1)})
    ax1.set_ylabel('Vp/Vs')
    ax1.set_xlabel('Depth km')
    ax1.set_xlim(min(HMoho), max(HMoho))
    ax1.set_ylim(min(KMoho), max(KMoho))
    #----------------------------------------------------------
    # Plot the receiver function witht the estimated times
    #----------------------------------------------------------
    ax2 = plt.subplot(gs[0, 1]) 
    if stack == True:
        baz=180
        rfdata = ((SequentialHKResult['RFMohoStack']) * g[0]) + baz
        major_ticks_x = np.arange(-10, 41, 5)                                           
        minor_ticks_x = np.arange(-10, 41, 1) 
        major_ticks_y = np.arange(0, 361, 45)                                              
        minor_ticks_y = np.arange(-30, 390, 5)
        ax2.set_xticks(major_ticks_x)                                                       
        ax2.set_xticks(minor_ticks_x, minor=True)                                           
        ax2.set_yticks(major_ticks_y)                                                       
        ax2.set_yticks(minor_ticks_y, minor=True)                                  
        ax2.set_xlim(-5, 30)
        ax2.set_ylim(-29, 390)
        ax2.plot(t, rfdata, "k-", lw=0.5)
        ax2.fill_between(t, baz, rfdata, where=rfdata > baz, 
                            facecolor='red', alpha = 0.25)
        ax2.fill_between(t, baz, rfdata, where=rfdata <= baz, 
                            facecolor='blue', alpha = 0.25)
        mtransforms.blended_transform_factory(ax2.transData, ax2.transAxes)
        #----------------------------------------------------------
        # Plot H-K sequential formula times
        #----------------------------------------------------------
        ampgain = g[1]
        ax2.plot([tpsMoho, tpsMoho], [baz+ampgain, baz-ampgain], 'g-', lw=1.5, label="Moho")
        ax2.plot([tpppsMoho, tpppsMoho], [baz+ampgain, baz-ampgain], 'g-', lw=1.5)
        ax2.plot([tpspsMoho, tpspsMoho], [baz+ampgain, baz-ampgain], 'g-', lw=1.5)
        filename = "%s/%s_%s_%s.%s"%(savepath, staname, comp, 'Moho', format)
        plt.savefig(filename , format=format, transparent=False,\
        dpi=250, bbox_inches = 'tight', pad_inches=0.1)
        plt.show()
        plt.close('all')
    else:
        rfstreamMoho= SequentialHKResult['RFMoho']
        gcarcarr = np.zeros(len(rfstreamMoho))
        for rf in rfstreamMoho:        
            delta = rf.stats.delta
            baz = rf.stats.back_azimuth
            b = rf.stats.starttime - rf.stats.onset
            tpsMoho = rf.stats.tpsMoho
            tpppsMoho = rf.stats.tpppsMoho
            tpspsMoho = rf.stats.tpspsMoho
            rfdata = ((rf.data) * g[0]) + baz
            t = (np.arange(0, len(rfdata), 1) * delta) + b
            major_ticks_x = np.arange(-10, 41, 5)                                           
            minor_ticks_x = np.arange(-10, 41, 1) 
            major_ticks_y = np.arange(0, 361, 45)                                              
            minor_ticks_y = np.arange(-30, 390, 5)
            ax2.set_xticks(major_ticks_x)                                                       
            ax2.set_xticks(minor_ticks_x, minor=True)                                           
            ax2.set_yticks(major_ticks_y)                                                       
            ax2.set_yticks(minor_ticks_y, minor=True)                                  
            ax2.set_xlim(-5, 30)
            ax2.set_ylim(-29, 390)
            ax2.plot(t, rfdata, "k-", lw=0.5)
            ax2.fill_between(t, baz, rfdata, where=rfdata > baz, 
                                facecolor='red', alpha = 0.25)
            ax2.fill_between(t, baz, rfdata, where=rfdata <= baz, 
                                facecolor='blue', alpha = 0.25)
            mtransforms.blended_transform_factory(ax2.transData, ax2.transAxes)
            #----------------------------------------------------------
            # Plot H-K sequential formula times
            #----------------------------------------------------------
            ampgain = g[1]
            ax2.plot([tpsMoho, tpsMoho], [baz+ampgain, baz-ampgain], 'g-', lw=1.5, label="Moho")
            ax2.plot([tpppsMoho, tpppsMoho], [baz+ampgain, baz-ampgain], 'g-', lw=1.5)
            ax2.plot([tpspsMoho, tpspsMoho], [baz+ampgain, baz-ampgain], 'g-', lw=1.5)
        #----------------------------------------------------------
        # Highlight specifiec time range on RF plot
        #----------------------------------------------------------
        ax2.axvspan(tpsMoho-0.5, tpsMoho+0.5, alpha=0.5, color='lightgreen')
        ax2.axvspan(tpppsMoho-0.5, tpppsMoho+0.5, alpha=0.5, color='lightgreen')
        ax2.axvspan(tpspsMoho-0.5, tpspsMoho+0.5, alpha=0.5, color='lightgreen')
        ax2.set_xlabel("Time (sec)")
        ax2.set_ylabel("Back-Azimuth (deg)")    
        #----------------------------------------------------------
        # Plot the receiver function witht the estimated times
        #----------------------------------------------------------
        ax3 = plt.subplot(gs[0, 2])
        gcarcarr = np.zeros(len(rfstreamMoho))   
        for rf, j in zip(rfstreamMoho, np.arange(len(rfstreamMoho))):        
            delta = rf.stats.delta
            gcarc = rf.stats.distance
            gcarcarr[j] = gcarc
            b = rf.stats.starttime - rf.stats.onset
            tpsMoho = rf.stats.tpsMoho
            tpppsMoho = rf.stats.tpppsMoho
            tpspsMoho = rf.stats.tpspsMoho
            rfdata = ((rf.data) * g[2]) + gcarc
            t = (np.arange(0, len(rfdata), 1) * delta) + b    
            major_ticks_x = np.arange(-10, 41, 5)                                             
            minor_ticks_x = np.arange(-10, 41, 1) 
            major_ticks_y = np.arange(30, 110, 10)                                              
            minor_ticks_y = np.arange(30, 111, 5)
            ax3.set_xticks(major_ticks_x)                                                       
            ax3.set_xticks(minor_ticks_x, minor=True)                                           
            ax3.set_yticks(major_ticks_y)                                                       
            ax3.set_yticks(minor_ticks_y, minor=True)                                          
            #----------------------------------------------------------
            # Plot grid on the figure
            #----------------------------------------------------------                                                               
            ax3.set_xlim(-5, 30)
            ax3.set_ylim(25,110)
            ax3.plot(t, rfdata, "k-", lw=0.5)
            ax3.fill_between(t, gcarc, rfdata, where=rfdata > gcarc, 
                                facecolor='red', alpha = 0.25)
            ax3.fill_between(t, gcarc, rfdata, where=rfdata <= gcarc, 
                                facecolor='blue', alpha = 0.25)
            mtransforms.blended_transform_factory(ax3.transData, ax3.transAxes)
            #----------------------------------------------------------
            # Plot H-K sequential formula times
            #----------------------------------------------------------
            ampgain = g[3]
            ax3.plot([tpsMoho, tpsMoho], [gcarc+ampgain, gcarc-ampgain], 'g-', lw=1.5, label="Moho")
            ax3.plot([tpppsMoho, tpppsMoho], [gcarc+ampgain, gcarc-ampgain], 'g-', lw=1.5)
            ax3.plot([tpspsMoho, tpspsMoho], [gcarc+ampgain, gcarc-ampgain], 'g-', lw=1.5)
        #----------------------------------------------------------
        # Highlight specifiec time range on RF plot
        #----------------------------------------------------------
        ax3.axvspan(tpsMoho-0.5, tpsMoho+0.5, alpha=0.5, color='lightgreen')
        ax3.axvspan(tpppsMoho-0.5, tpppsMoho+0.5, alpha=0.5, color='lightgreen')
        ax3.axvspan(tpspsMoho-0.5, tpspsMoho+0.5, alpha=0.5, color='lightgreen')        
        ax3.set_xlabel("Time (sec)")
        ax3.set_ylabel("Distance (deg)")
        filename = "%s/%s_%s_%s.%s"%(savepath, staname, comp, 'Moho', format)
        plt.savefig(filename , format=format, transparent=False,\
            dpi=250, bbox_inches = 'tight', pad_inches=0.1)
        plt.show()


"""
.. module:: ReverbFilter
        :synopsis: Resonance Filter for receiver function data (Removes sediment reverberation effect)
.. moduleauthor:: Stephen Akinremi <s.akinremi@utwente.nl> Islam Fadel <i.e.a.m.fadel@utwente.nl> (October 2023)
"""

import numpy as np
from scipy import signal
from scipy.signal import argrelextrema, argrelmin, argrelmax
import math
import cmath
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.transforms as mtransforms
from scipy.fftpack import fft, ifft
from matplotlib.legend_handler import HandlerLine2D as HL
import seaborn
seaborn.set_style("darkgrid", {"axes.facecolor": ".85"})

#----------------------------------------------------------
def getamp(rfdata, tarray, t):
    """
    Get the amplitude of the receiver function at a specific time

    :param rfdata: receiver function data
    :type rfdata: numpy array
    :param tarray: time array
    :type tarray: numpy array
    :param t: time to get the amplitude
    :type t: float
    :return: amplitude of the receiver function at time t
    """
    amp = rfdata[(np.abs(tarray - t).argmin())]
    return amp
#----------------------------------------------------------
def Resonance_Filt(rfstream):
    """
    Resonance Filter for receiver function data (Removes sediment reverberation effect)

    :param rfstream: receiver function data
    :type rfstream: obspy stream

    :return: A dictionary of stacked receiver function, Filtered receiver function, 
            autocorrelation of the data, resonance filter, time lag (2 way travel time of the sediment reverbration), 
            and the strength of the sediment reverbration (r)
    """
    delta = rfstream[0].stats.delta
    staname=rfstream[0].stats.station
    comp =  rfstream[0].stats.channel
    b = rfstream[0].stats.starttime - rfstream[0].stats.onset
    t = (np.arange(0, len(rfstream[0]), 1) * delta) + b
    Dt=rfstream[0].stats.delta
    StackedRF=rfstream.stack()
    StackedData=StackedRF[0].data
    tzero=np.where(t==0)[0]
    tzero=int(tzero)  
    start=tzero
    rf=StackedData[start::] 
    time = t[start::]
    Dt=delta
    N=len(rf)
    fmax=1/(2*Dt)
    df=fmax/(N/2)
    #---------------------------------------------
    # Build the frequency vector
    N2=N/2
    #---------------------------------------------
    column1=np.arange(0,N2+1)    
    column2=np.arange(-N2+1,-1) 
    #---------------------------------------------
    # Join the two vectors
    full_column = np.concatenate((column1,column2))
    #---------------------------------------------
    f=df*full_column
    Nf = N/2+1
    dw=2*np.pi*df
    w=dw*full_column
    #----------------------------------------------------------------------------------------------------------------------
    # Do Auto-correlation
    #----------------------------------------------------------------------------------------------------------------------
    D=rf
    D=D-np.mean(D)
    D=signal.detrend(D, axis=-1, type='linear')
    #------------------------
    autoc=signal.correlate(D,D,mode='full')#,method='auto')
    autoc=autoc/max(autoc)
    autoc=autoc[N-1:2*N-1]
    #----------------------------------------------------------------------------------------------------------------------
    # Find time lag and amplitude of the autocorrelation
    #----------------------------------------------------------
    local_min=argrelmin(autoc)
    index=(local_min[0][0])
    r=(autoc[index])
    r=math.sqrt(r**2)
    tlag= time[index]
    #------------------------------------------------------------------------------
    # Build Filter
    #------------------------------------------------------------------------------
    resonanceflt = []
    for i in range(0, len(w)):
        flt= (1 + (r)*cmath.exp((-1j)*(w[i])*tlag))
        resonanceflt.append(flt)
    #------------------------------------------------------------------------------
    # Filter the Data with a Resonance Filter (Fourier Transform)
    F_trans=fft(rf)
    Filtered=F_trans*resonanceflt
    F_inv=ifft(Filtered)
    #Use Real Part of the Inverse Fourier Transform
    filtered_rf=np.real(F_inv)  #real part of the inverse fourier transform
    #------------------------------------------------------------------------------
    FltResults={'rf':rf, 'filteredrf': filtered_rf, 'resonancefilter':resonanceflt, 'time':time, 'autoc':autoc, 'r':r, 'tlag':tlag, 'staname':staname, 'delta':delta}
    return FltResults
#----------------------------------------------------------
def plotfiltrf(FilteredRF, savepath, format = 'jpg'):
    """
    Plot the filtered receiver function, autocorrelation, and the resonance filter

    :param FilteredRF: Dictionary of results from the resonance filtering method (function: Resonance_Filt)
    :type FilteredRF: dict
    :param savepath: path to save the plots
    :type savepath: str
    :param format: format of the plot (default: jpg)
    :type format: str
    """
    rf=FilteredRF['rf']
    filtered_rf=FilteredRF['filteredrf']
    resonanceflt=FilteredRF['resonancefilter']
    time=FilteredRF['time']
    autoc=FilteredRF['autoc']
    staname=FilteredRF['staname']
    # Plot Autocorrelation
    plt.plot(time,autoc)
    plt.xlabel('Time (s)')
    plt.ylabel('Amplitude')
    plt.title('Autocorrelation of %s'%staname)
    filename = "%s/%s.%s"%(savepath, 'Autocorrelation', format)
    plt.savefig(filename , format=format, transparent=False,\
            dpi=250, bbox_inches = 'tight', pad_inches=0.1)
    plt.show()
    plt.close("all")
    # Plot Resonance Filter
    plt.plot(time, resonanceflt)
    plt.xlabel('Time (s)')
    # plt.ylabel('Amplitude')
    plt.title('Resonance Filter of %s'%staname)
    filename = "%s/%s.%s"%(savepath, "Resonance_Filter", format)
    plt.savefig(filename , format=format, transparent=False,\
            dpi=250, bbox_inches = 'tight', pad_inches=0.1)
    plt.show()
    plt.close("all")
    # Plot RF and Filtered RF
    plt.plot(time, rf, label = 'RF', color = 'black')
    plt.plot(time, filtered_rf, label = 'Filtered RF', color = 'red')
    plt.xlabel('Time (s)')
    plt.ylabel('Amplitude')
    plt.title('RF and Filtered RF of %s'%staname)
    plt.legend()
    filename = "%s/%s.%s"%(savepath, "RF_FilteredRF", format)
    plt.savefig(filename , format=format, transparent=False,\
            dpi=250, bbox_inches = 'tight', pad_inches=0.1)
    plt.show()
    plt.close("all")
#----------------------------------------------------------





# Copyright 2013-2016 Tom Eulenfeld, MIT license
"""
Tests for batch module.
"""
from glob import glob
import unittest
import os
from pkg_resources import load_entry_point
import sys
import warnings

import matplotlib
matplotlib.use('Agg')

from rfsed.rf.batch import init_data, run_cli as script
from rfsed.rf.tests.util import quiet, tempdir
try:
    import obspyh5
except ImportError:
    obspyh5 = None


def substitute(old, new):
    fname = 'conf.json'
    with open(fname) as f:
        text = f.read()
    text = text.replace(old, new)
    with open(fname, 'w') as f:
        f.write(text)


def test_format(testcase, format):
    join = os.path.join

    # QHD
    with tempdir():
        script(['create', '-t'])
        substitute('#"format": "Q"', '"format": "%s"' % format)
        script(['data', 'data'])
        script(['calc', 'moveout', 'data', 'mout1'])
        script(['data', 'calc', 'datarf'])
        script(['moveout', 'datarf', 'mout2'])
        script(['stack', 'mout1', 'stack'])
        script(['profile', 'mout1', 'profile'])
        if format in ('Q', 'SAC'):
            patterns = [join('data', '*', '*'), join('mout1', '*', '*'),
                        join('mout2', '*', '*'), join('stack', '*'),
                        join('profile', '*') if format == 'SAC' else
                        'profile*.*']
        else:
            patterns = ['data.h5', 'mout1.h5', 'mout2.h5', 'stack.h5',
                        'profile.h5']
        nums = [len(glob(p)) for p in patterns]
        nums2 = {'Q': [14, 14, 14, 2, 2],
                 'SAC': [21, 21, 21, 3, 6],
                 'H5': [1, 1, 1, 1, 1]}
        testcase.assertEqual(nums, nums2[format])
        if format in ('Q', 'H5'):
            script(['convert', 'mout1', 'mout_SAC', 'SAC'])
        if format in ('H5', 'SAC'):
            script(['convert', 'mout1', 'mout_Q', 'Q'])
        if obspyh5 and format in ('Q', 'SAC'):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                script(['convert', 'mout1', 'mout_H5', 'H5'])
        script(['plot', 'mout1', 'plot'])
        script(['plot-profile', 'profile', 'plot_profile'])
        testcase.assertEqual(len(glob(join('plot', '*.pdf'))), 3)
        testcase.assertEqual(len(glob(join('plot_profile', '*.pdf'))), 3)


class BatchTestCase(unittest.TestCase):
    """
    For now batch tests are not run on windows.
    See https://travis-ci.org/trichter/rf/jobs/589375488 for failures.
    The batch module has a low priority.
    """
    def setUp(self):
        # turn off progressbar
        import rfsed.rf.batch
        rfsed.rf.batch.tqdm = lambda: None

    @unittest.skipIf(sys.platform.startswith("win"), "fails on Windows")
    def test_entry_point(self):
        ep_script = load_entry_point('rfsed', 'console_scripts', 'rf')
        try:
            with quiet():
                ep_script(['-h'])
        except SystemExit:
            pass

    @unittest.skipIf(sys.platform.startswith("win"), "fails on Windows")
    def test_batch_command_interface_Q(self):
        test_format(self, 'Q')

    @unittest.skipIf(sys.platform.startswith("win"), "fails on Windows")
    def test_batch_command_interface_SAC(self):
        test_format(self, 'SAC')

    @unittest.skipIf(obspyh5 is None, 'obspyh5 not installed')
    @unittest.skipIf(sys.platform.startswith("win"), "fails on Windows")
    def test_batch_command_interface_H5(self):
        test_format(self, 'H5')

    def test_plugin_option(self):
        f = init_data('plugin', plugin='rfsed.rf.tests.test_batch : gw_test')
        self.assertEqual(f(nework=4, station=2), 42)


def gw_test(**kwargs):
    return 42


def suite():
    return unittest.makeSuite(BatchTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



import unittest
from rfsed import ExtractEq
from rfsed.util import raw_data_example, save_tests, catalog_example

class TestExtractEq(unittest.TestCase):
    def test_ExtractEq(self):
        datapath=raw_data_example()
        staname='NE301'
        stalat=6.183432
        stalon=53.48334
        savepath=save_tests()
        Request_window=[-50, 150]
        filename = savepath + 'eq_data.h5'
        catalog = catalog_example()
        ExtractEq.ExtractEq(datapath=datapath, catalog=catalog, filename=filename, stalat=stalat, stalon=stalon, Request_window=Request_window)
        
        
def suite():
    return unittest.makeSuite(TestExtractEq, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')


# Copyright 2013-2019 Tom Eulenfeld, MIT license
"""
All objects and functions below are after the Tom Eulenfeld's
rf project License below.

The MIT License (MIT)

Copyright (c) 2013-2019 Tom Eulenfeld

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
"""
Frequency and time domain deconvolution.
"""
from copy import copy
import numpy as np
from numpy import max, pi
from scipy.fftpack import fft, ifft, next_fast_len
from scipy.signal import correlate, detrend
from rfsed.rf.util import _add_processing_info


def __find_nearest(array, value):
    """http://stackoverflow.com/a/26026189"""
    idx = np.searchsorted(array, value, side='left')
    expr = np.abs(value - array[idx - 1]) < np.abs(value - array[idx])
    if idx > 0 and (idx == len(array) or expr):
        return idx - 1
    else:
        return idx


@_add_processing_info
def deconvolve(stream, method='time', func=None,
               source_components='LZ', response_components=None,
               winsrc='P', **kwargs):
    """
    Deconvolve one component of a stream from other components.

    The deconvolutions are written to the data arrays of the stream. To keep
    the original data use the copy method of Stream.
    The stats dictionaries of the traces inside stream must have an 'onset'
    entry with a `~obspy.core.utcdatetime.UTCDateTime` object.
    This will be used for determining the data windows.

    :param stream: stream including responses and source
    :param method:
        'time' -> use time domain deconvolution, see `deconv_time()`,\n
        'waterlevel' -> use frequency domain deconvolution with water level, see `deconv_waterlevel()`\n
        'iterative' -> use iterative time domain deconvolution, see `deconv_iterative()`\n
        'multitaper' -> use frequency domain multitaper deconvolution, see `deconv_multitaper()`\n
        'func' -> user defined function (func keyword)
    :param func: Custom deconvolution function with the following signature

             def deconv_custom(rsp: RFStream, src: RFTrace, tshift=10,
                               \*\*other_kwargs_possible) -> RFStream:

    :param source_components: names of components identifying the source traces,
        e.g. 'LZ' for P receiver functions and 'QR' for S receiver functions
    :param response_components: names of components identifying the response
        traces (default None: all traces are used as response)
    :type winsrc: tuple (start, end, taper)
    :param winsrc: data window for source function, in seconds relative to
        onset. The function will be cosine tapered at both ends (seconds).\n
        winsrc can also be a string ('P' or 'S'). In this case the function
        defines a source time window appropriate for this type of receiver
        function and deconvolution method (see source code for details).
    :param \*\*kwargs: other kwargs are passed to the underlying deconvolution
        functions

    .. note::
        If parameter normalize is not present in kwargs and source component is
        not excluded from the results by response_components, results will be
        normalized such that the maximum of the deconvolution of the trimmed
        and tapered source from the untouched source is 1. If the source is
        excluded from the results, the normalization will performed against
        the first trace in results.

    .. note::
        If multitaper deconvolution is used and a stream of (pre-event) noise
        is not present in kwargs, noise will be sampled from the data in a
        pre-event window whose length depends on the trace length prior to the
        onset time.
    """
    if method == 'freq':
        method = 'waterlevel'
    if method not in ('time', 'waterlevel', 'iterative', 'multitaper', 'func'):
        raise NotImplementedError()
    # identify source and response components
    src = [tr for tr in stream if tr.stats.channel[-1] in source_components]
    if len(src) != 1:
        msg = 'Invalid number of source components. %d not equal to one.'
        raise ValueError(msg % len(src))
    src = src[0]
    rsp = [tr for tr in stream if response_components is None or
           tr.stats.channel[-1] in response_components]
    if 'normalize' not in kwargs and src in rsp:
        kwargs['normalize'] = rsp.index(src)
    if not 0 < len(rsp) < 4:
        msg = 'Invalid number of response components. %d not between 0 and 4.'
        raise ValueError(msg % len(rsp))

    sr = src.stats.sampling_rate
    # shift onset to time of nearest data sample to circumvent complications
    # for data with low sampling rate and method='time'
    idx = __find_nearest(src.times(), src.stats.onset - src.stats.starttime)
    src.stats.onset = onset = src.stats.starttime + idx * src.stats.delta
    for tr in rsp:
        tr.stats.onset = onset
    # define default time windows
    lenrsp_sec = src.stats.endtime - src.stats.starttime
    onset_sec = onset - src.stats.starttime
    if winsrc == 'P' and method in ('time', 'multitaper', 'func'):
        winsrc = (-10, 30, 5)
    elif winsrc == 'S' and method in ('time','multitaper', 'func'):  # TODO: test this
        winsrc = (-10, 30, 5)
    elif winsrc == 'P' and method == 'iterative':
        winsrc = (-onset_sec, lenrsp_sec - onset_sec, 0)
    elif winsrc == 'S' and method == 'iterative':  # TODO: test this
        winsrc = (-onset_sec, lenrsp_sec - onset_sec, 0)
    elif winsrc == 'P' and method == 'waterlevel':
        winsrc = (-onset_sec, lenrsp_sec - onset_sec, 5)
    elif winsrc == 'S' and method == 'waterlevel':
        winsrc = (-10, lenrsp_sec - onset_sec, 5)
#    winsrc = list(winsrc)
#    if winsrc[0] < -onset_sec:
#        winsrc[0] = -onset_sec
#    if winsrc[1] > lenrsp_sec - onset_sec:
#        winsrc[1] = lenrsp_sec - onset_sec
    # prepare source and response list
    if src in rsp:
        src = src.copy()
    src.trim(onset + winsrc[0], onset + winsrc[1], pad=True, fill_value=0.)
    src.taper(max_percentage=None, max_length=winsrc[2])
    tshift = -winsrc[0]
    if method == 'time':
        shift = int(round(tshift * sr - len(src) // 2))
        rsp_data = [tr.data for tr in rsp]
        rf_data = deconv_time(rsp_data, src.data, shift,  **kwargs)
        for i, tr in enumerate(rsp):
            tr.data = rf_data[i].real
    elif method == 'waterlevel':
        rsp_data = [tr.data for tr in rsp]
        rf_data = deconv_waterlevel(rsp_data, src.data, sr, tshift=tshift,
                                    **kwargs)
        for i, tr in enumerate(rsp):
            tr.data = rf_data[i].real
    elif method == 'iterative':
        rsp_data = [tr.data for tr in rsp]
        rf_data, nit, _ = deconv_iterative(rsp_data, src.data, sr,
                                           tshift=tshift, **kwargs)
        for i, tr in enumerate(rsp):
            tr.data = rf_data[i].real
            tr.stats['iterations'] = nit[i]
    elif method == 'multitaper':
        noise = kwargs.pop('noise',None)
        if noise is None:  # no kwarg, grab from pre-event time series
            onset_rsp = rsp[0].stats.onset - rsp[0].stats.starttime
            noise = stream.copy().trim2(-onset_rsp,-5,'onset') # NOTE window length will vary
        # noise is not None (kwarg provided), noise should be Stream() or RFStream()
        # so now we grab that data and make a list of arrays
        nse_data = [tr.data for tr in noise if response_components is None or
                    tr.stats.channel[-1] in response_components]
        rsp_data = [tr.data for tr in rsp]
        rf_data = deconv_multitaper(rsp_data, src.data, nse_data, sr, -tshift,
                                    **kwargs)
        for i, tr in enumerate(rsp):
            tr.data = rf_data[i].real
    else:
        rsp = func(stream.__class__(rsp), src, tshift=tshift, **kwargs)
    return stream.__class__(rsp)


def __get_length(rsp_list):
    if isinstance(rsp_list, (list, tuple)):
        rsp_list = rsp_list[0]
    return len(rsp_list)


def deconv_waterlevel(rsp_list, src, sampling_rate, waterlevel=0.05, gauss=0.5,
                      tshift=10., length=None, normalize=0, nfft=None,
                      return_info=False):
    """
    Frequency-domain deconvolution using waterlevel method.

    Deconvolve src from arrays in rsp_list.

    :param rsp_list: either a list of arrays containing the response functions
        or a single array
    :param src: array with source function
    :param sampling_rate: sampling rate of the data
    :param waterlevel: waterlevel to stabilize the deconvolution
    :param gauss: Gauss parameter (standard deviation) of the
        Gaussian Low-pass filter,
        corresponds to cut-off frequency in Hz for a response value of
        exp(-0.5)=0.607.
    :param tshift: delay time 0s will be at time tshift afterwards
    :param nfft: explicitely set number of samples for the fft
    :param length: number of data points in results, optional
    :param normalize: normalize all results so that the maximum of the trace
        with supplied index is 1. Set normalize to 'src' to normalize
        for the maximum of the prepared source. Set normalize to None for no
        normalization.
    :param return_info: return additionally a lot of different parameters in a
        dict for debugging purposes

    :return: (list of) array(s) with deconvolution(s)
    """
    if length is None:
        length = __get_length(rsp_list)
    N = length
    if nfft is None:
        nfft = next_fast_len(N)
    dt = 1. / sampling_rate
    ffilt = _phase_shift_filter(nfft, dt, tshift)
    if gauss is not None:
        ffilt = _gauss_filter(dt, nfft, gauss, waterlevel=-700) * ffilt
    spec_src = fft(src, nfft)
    spec_src_conj = np.conjugate(spec_src)
    spec_src_water = np.abs(spec_src * spec_src_conj)
    spec_src_water = np.maximum(
        spec_src_water, max(spec_src_water) * waterlevel)

    if normalize == 'src':
        spec_src = ffilt * spec_src * spec_src_conj / spec_src_water
        rf_src = ifft(spec_src, nfft)[:N]
        norm = 1 / max(rf_src)
        rf_src = norm * rf_src

    flag = False
    if not isinstance(rsp_list, (list, tuple)):
        flag = True
        rsp_list = [rsp_list]
    rf_list = [ifft(ffilt * fft(rsp, nfft) * spec_src_conj / spec_src_water,
                    nfft)[:N] for rsp in rsp_list]
    if normalize not in (None, 'src'):
        norm = 1. / max(rf_list[normalize])
    if normalize is not None:
        for rf in rf_list:
            rf *= norm
    if return_info:
        if normalize not in (None, 'src'):
            spec_src = ffilt * spec_src * spec_src_conj / spec_src_water
            rf_src = ifft(spec_src, nfft)[:N]
            norm = 1 / max(rf_src)
            rf_src = norm * rf_src
        info = {'rf_src': rf_src, 'rf_src_conj': spec_src_conj,
                'spec_src_water': spec_src_water,
                'freq': np.fft.fftfreq(nfft, d=dt),
                'gauss': ffilt, 'norm': norm, 'N': N, 'nfft': nfft}
        return rf_list, info
    elif flag:
        return rf
    else:
        return rf_list


def _add_zeros(a, numl, numr):
    """Add zeros at left and rigth side of array a"""
    return np.hstack([np.zeros(numl), a, np.zeros(numr)])


def _acorrt(a, num):
    """
    Not normalized auto-correlation of signal a.

    Sample 0 corresponds to zero lag time. Auto-correlation will consist of
    num samples. Correlation is performed in time domain with scipy.

    :param a: Data
    :param num: Number of returned data points
    :return: autocorrelation
    """
    return correlate(_add_zeros(a, 0, num - 1), a, 'valid')


def _xcorrt(a, b, num, zero_sample=0):
    """
    Not normalized cross-correlation of signals a and b.

    :param a,b: data
    :param num: The cross-correlation will consist of num samples.\n
        The sample with 0 lag time will be in the middle.
    :param zero_sample: Signals a and b are aligned around the middle of their
        signals.\n
        If zero_sample != 0 a will be shifted additionally to the left.
    :return: cross-correlation
    """
    if zero_sample > 0:
        a = _add_zeros(a, 2 * abs(zero_sample), 0)
    elif zero_sample < 0:
        a = _add_zeros(a, 0, 2 * abs(zero_sample))
    dif = len(a) - len(b) + 1 - num
    if dif > 0:
        b = _add_zeros(b, (dif + 1) // 2, dif // 2)
    else:
        a = _add_zeros(a, (-dif + 1) // 2, (-dif) // 2)
    return correlate(a, b, 'valid')


# Gives similar results as a deconvolution with Seismic handler,
# but SH is faster
def deconv_time(rsp_list, src, shift, spiking=1., length=None, normalize=0,
                solve_toeplitz='toeplitz'):
    """
    Time domain deconvolution.

    Deconvolve src from arrays in rsp_list.
    Calculate Toeplitz auto-correlation matrix of source, invert it, add noise
    and multiply it with cross-correlation vector of response and source.

    In one formula::

        RF = (STS + spiking*I)^-1 * STR

        N... length
            ( S0   S-1  S-2 ... S-N+1 )
            ( S1   S0   S-1 ... S-N+2 )
        S = ( S2   ...                )
            ( ...                     )
            ( SN-1 ...          S0    )
        R = (R0 R1 ... RN-1)^T
        RF = (RF0 RF1 ... RFN-1)^T
        S... source matrix (shape N*N)
        R... response vector (length N)
        RF... receiver function (deconvolution) vector (length N)
        STS = S^T*S = symmetric Toeplitz autocorrelation matrix
        STR = S^T*R = cross-correlation vector
        I... Identity

    :param rsp_list: either a list of arrays containing the response functions
        or a single array
    :param src: array of source function
    :param shift: shift the source by that amount of samples to the left side
        to get onset in RF at the desired time (negative -> shift source to the
        right side)\n
        shift = (middle of rsp window - middle of src window) +
        (0 - middle rf window)
    :param spiking: random noise added to autocorrelation (eg. 1.0, 0.1)
    :param length: number of data points in results
    :param normalize: normalize all results so that the maximum of the trace
        with supplied index is 1. Set normalize to None for no normalization.
    :param solve_toeplitz: Python function to use as a solver for Toeplitz
        system. Possible values are
        ``'toeplitz'`` for using the toeplitz package (default),
        ``'scipy'`` for using solve_toeplitz from the SciPy package, or
        a custom function.

    :return: (list of) array(s) with deconvolution(s)
    """
    if solve_toeplitz == 'toeplitz':
        try:
            from toeplitz import sto_sl
        except ImportError:
            from warnings import warn
            warn('Toeplitz import error. Fallback to slower Scipy version.')
            from scipy.linalg import solve_toeplitz
        else:
            def solve_toeplitz(a, b):
                """
                Solve linear system Ax=b for real symmetric Toeplitz matrix A.

                :param a: first row of Toeplitz matrix A
                :param b: vector b
                :return: x=A^-1*b
                """
                return sto_sl(np.hstack((a, a[1:])), b, job=0)
    elif solve_toeplitz == 'scipy':
        from scipy.linalg import solve_toeplitz
    if length is None:
        length = __get_length(rsp_list)
    flag = False
    RF_list = []
    STS = _acorrt(src, length)
    STS = STS / STS[0]
    STS[0] += spiking
    if not isinstance(rsp_list, (list, tuple)):
        flag = True
        rsp_list = [rsp_list]
    for rsp in rsp_list:
        STR = _xcorrt(rsp, src, length, shift)
        assert len(STR) == len(STS)
        RF = solve_toeplitz(STS, STR)
        RF_list.append(RF)
    if normalize is not None:
        norm = 1 / np.max(np.abs(RF_list[normalize]))
        for RF in RF_list:
            RF *= norm
    if flag:
        return RF
    else:
        return RF_list


def _gauss_filter(dt, nft, f0, waterlevel=None):
    """
    Gaussian filter with width f0

    :param dt: sample spacing in seconds
    :param nft: length of filter in points
    :param f0: Standard deviation of the Gaussian Low-pass filter,
        corresponds to cut-off frequency in Hz for a response value of
        exp(0.5)=0.607.
    :param waterlevel: waterlevel for eliminating very low values
        (default: no waterlevel)
    :return: array with Gaussian filter frequency response
    """
    f = np.fft.fftfreq(nft, dt)
    gauss_arg = -0.5 * (f/f0) ** 2
    if waterlevel is not None:
        gauss_arg = np.maximum(gauss_arg, waterlevel)
    return np.exp(gauss_arg)


def _phase_shift_filter(nft, dt, tshift):
    """
    Construct filter to shift an array to account for time before onset

    :param nft: number of points for fft
    :param dt: sample spacing in seconds
    :param tshift: time to shift by in seconds
    :return: shifted array
    """
    freq = np.fft.fftfreq(nft, d=dt)
    return np.exp(-2j * pi * freq * tshift)


def _apply_filter(x, filt):
    """
    Apply a filter defined in frequency domain to a data array

    :param x: array of data to filter
    :param filter: filter to apply in frequency domain,
        e.g. from _gauss_filter()
    :return: real part of filtered array
    """
    nfft = len(filt)
    xf = fft(x, n=nfft)
    xnew = ifft(xf*filt, n=nfft)
    return xnew.real


def deconv_iterative(rsp, src, sampling_rate, tshift=10, gauss=0.5, itmax=400,
                     minderr=0.001, mute_shift=False, normalize=0):
    """
    Iterative deconvolution.

    Deconvolve src from arrays in rsp.
    Iteratively construct a spike train based on least-squares minimzation of the
    difference between one component of an observed seismogram and a predicted
    signal generated by convolving the spike train with an orthogonal component
    of the seismogram.

    Reference: Ligorria, J. P., & Ammon, C. J. (1999). Iterative Deconvolution
    and Receiver-Function Estimation. Bulletin of the Seismological Society of
    America, 89, 5.

    :param rsp: either a list of arrays containing the response functions
        or a single array
    :param src: array with source function
    :param sampling_rate: sampling rate of the data
    :param tshift: delay time 0s will be at time tshift afterwards
    :param gauss: Gauss parameter (standard deviation) of the
        Gaussian Low-pass filter,
        corresponds to cut-off frequency in Hz for a response value of
        exp(-0.5)=0.607.
    :param itmax: limit on number of iterations/spikes to add
    :param minderr: stop iteration when the change in error from adding another
        spike drops below this threshold
    :param mute_shift: Mutes all samples at beginning of trace
        (lenght given by time shift).
        For `len(src)==len(rsp)` this mutes all samples before the onset.
    :param normalize: normalize all results so that the maximum of the trace
        with the supplied index is 1. Set normalize to None for no normalization.

    :return: (list of) array(s) with deconvolution(s)
    """

    ncomp = len(rsp)    # number of components we're looping over here
    dt = 1 / sampling_rate
    RF_out = []
    it_out = []  # number of iterations each component uses
    rms_out = []

    for c in range(ncomp):  # loop over the responses
        r0 = rsp[c]
        nfft = next_fast_len(2 * len(r0))
        rms = np.zeros(itmax)    # to store rms
        p0 = np.zeros(nfft)      # and rf for this component iteration

        gaussF = _gauss_filter(dt, nfft, gauss)  # construct and apply gaussian filter
        r_flt = _apply_filter(r0, gaussF)
        s_spec_flt = fft(src, nfft) * gaussF  # spectrum of source
        powerS = np.sum(ifft(s_spec_flt).real ** 2)  # power in the source for scaling
        powerR = np.sum(r_flt**2)  # power in the response for scaling
        rem_flt = copy(r_flt)  # thing to subtract from as spikes are added to p0

        it = 0
        sumsq_i = 1
        d_error = 100*powerR + minderr
        mute_min = len(r0)
        mute_max = nfft if mute_shift else nfft - int(tshift*sampling_rate)
        while np.abs(d_error) > minderr and it < itmax:  # loop iterations, add spikes
            rs = ifft(fft(rem_flt) * np.conj(s_spec_flt)).real  # correlate (what's left of) the num & demon, scale
            rs = rs / powerS / dt  # scale the correlation
            rs[mute_min:mute_max] = 0
            i1 = np.argmax(np.abs(rs))  # index for getting spike amplitude
            # note that ^abs there means negative spikes are allowed
            p0[i1] = p0[i1] + rs[i1]  # add the amplitude of the spike to our spike-train RF
            p_flt = ifft(fft(p0) * s_spec_flt).real * dt  # convolve with source

            rem_flt = r_flt - p_flt  # subtract spike estimate from source to see what's left to model
            sumsq = np.sum(rem_flt**2) / powerR
            rms[it] = sumsq   # save rms
            d_error = 100 * (sumsq_i - sumsq)  # check change in error as a result of this iteration

            sumsq_i = sumsq     # update rms
            it = it + 1         # and add one to the iteration count
        # once we get out of the loop:
        shift_filt = _phase_shift_filter(nfft, dt, tshift)
        p_flt = _apply_filter(p0, gaussF * shift_filt)
        RF_out.append(p_flt[:len(r0)])  # save the RF for output
        it_out.append(it)
        rms_out.append(rms)

    if normalize is not None:
        norm = 1 / np.max(np.abs(RF_out[normalize]))
        for RF in RF_out:
            RF *= norm

    return RF_out, it_out, rms_out


def deconv_multitaper(rsp, src, nse, sampling_rate, tshift, gauss=0.5,
                      K=3, tband=4, T=10, olap=0.75, normalize=0):
    """
    Multitaper frequency domain deconvolution

    Deconvolve src from arrays in rsp.

    Based on mtdecon.f by G. Helffrich with some slight modifications

    References:
    Helffrich, G (2006). Extended-time multitaper frequency domain
    cross-correlation receiver function estimation. Bulletin of the
    Seismological Society of America, 96 (1).
    Shibutani, T., Ueno, T., & Hirahara, K. (2008). Improvement in the
    Extended-Time Multitaper Receiver Function Estimation Technique.
    Bulletin of the Seismological Society of America, 98 (2).

    :param rsp: a list of arrays containing the response functions
    :param src: array of source function
    :param nse: a list of arrays containing samples of pre-event noise
    :param sampling_rate: sampling rate of the data
    :param tshift: shift the source by that amount of samples to the left side
        to get onset in RF at the desired time (src time window length pre-onset)
    :param gauss: Gauss parameter (standard deviation) of the
        Gaussian Low-pass filter,
        corresponds to cut-off frequency in Hz for a response value of
        exp(-0.5)=0.607.
    :param K: number of Slepian tapers to use (default: 3)
    :param tband: time-bandwidth product (default: 4)
    :param T: time length of taper window in seconds (default: 10)
    :param olap: window overlap between 0 and 1 (default: 0.75)
    :param normalize: normalize all results so that the maximum of the trace
        with supplied index is 1. Set normalize to None for no normalization.

    :return: (list of) array(s) with deconvolution(s)
    """
    try:
        import mtspec as mt
    except ImportError as ex:
        msg = 'mtspec package is needed for multitaper deconvolution'
        raise ImportError(msg) from ex

    # check src trace length < rsp trace length:
    assert len(src) < len(rsp[0]), 'source wavelet must be shorter than response'
    assert len(nse[0]) <= len(rsp[0]), 'noise should not be longer than response'

    nft = len(rsp[0])  # length of final arrays
    dt = 1./sampling_rate  # sample spacing

    # calculate multitapers with mtspec
    ntap = int(round(T/dt))  # #points in each taper

    tap, el, _ = mt.multitaper.dpss(ntap, tband, K); tap = tap.T
    if K >= 2:
        tap[1] = -tap[1]  # adjust to match sign convention
    if K >= 3:
        tap[2] = -tap[2]
    if K > 3:
        print('Warning: taper signs may need adjustment for K>3')

    ofac = 1 / (1 - olap)  # calculate overlap factor

    sfft = np.zeros((K,nft), dtype=np.complex)  # array for holding freq-domain source estimate
    src = detrend(src, type='constant')  # demean and detrend source
    src = detrend(src, type='linear')

    # window and taper the source
    nwav = len(src)   # number of nonzero points for the source
    nwin = int(ofac*nwav/ntap)  # number of windows needed to cover the source

    src_pad = np.zeros(nft)
    src_pad[:nwav] = src
    for j in range(nwin):
        js = int(j*ntap/ofac)  # get indices for overlapping taper window
        je = int(js + ntap)
        for k in range(K):
            bit = np.zeros(nft)
            bit[js:je] = tap[k] # insert window taper into full-length zeros
            bit = bit*src_pad   # window/taper the padded source trace
            bit = fft(bit)      # transform tapered bit to frequency domain
            sfft[k] = sfft[k] + bit  # sum to total freq-domain source estimate

    # multiply by factor to account for short wavelet relative to fft
    fac = nwav/float(nwin*nft)
    sfft = fac*sfft

    # loop response components and do a similar window/taper/transform thing
    ncomp = len(rsp)
    RF_out = np.zeros((ncomp, nft))
    for c in range(ncomp):
        dat = rsp[c]; nos = nse[c]  # pick out component
        dfft = np.zeros((K,nft), dtype=np.complex)  # arrays for storing freq-domain estimates
        nfft = np.zeros((K,nft), dtype=np.complex)

        # window, taper, and transform the noise
        nos = detrend(nos, type='constant')
        nos = detrend(nos, type='linear')
        nos_pad = np.zeros(nft)
        nos_pad[:len(nos)] = nos
        nwin = int(ofac*len(nos)/ntap)
        for j in range(nwin):
            js = int(j*ntap/ofac)
            je = int(js + ntap)
            if je < nft:
                for k in range(K):
                    bit = np.zeros(nft)
                    bit[js:je] = tap[k]
                    bit = bit*nos_pad
                    bit = fft(bit)
                    nfft[k] = nfft[k] + bit

        fac = nwav/float(nwin*nft)  # scale for noise trace length
        nfft = nfft*fac

        # calculate power in the noise window
        s0 = np.zeros(nft)
        for k in range(K):
            s0 = s0 + (np.real(nfft[k])**2 + np.imag(nfft[k])**2)/el[k]

        # window, taper, and transform the response traces
        dat = detrend(dat, type='constant')
        dat = detrend(dat, type='linear')
        nwin = int(ofac*nft/ntap)-1
        for j in range(nwin):
            js = int(j*ntap/ofac)
            je = int(js + ntap)
            if je < nft:
                for k in range(K):
                    bit = np.zeros(nft)
                    bit[js:je] = tap[k]
                    bit = bit*dat
                    bit = fft(bit)
                    dfft[k] = dfft[k] + bit

        fac = float(nft)/float(nwin*nft)
        dfft = dfft*fac

        # deconvolve
        num = np.zeros(nft)
        denom = np.zeros(nft)
        for k in range(K):
            num = num + sfft[k]*np.conj(dfft[k])
            denom = denom + sfft[k]*np.conj(sfft[k])
        recF = num/(denom + s0)

        # time-shift for onset and apply gaussian lowpass
        if gauss is not None:
            recF = recF * _gauss_filter(dt, nft, gauss)
        recF = recF * _phase_shift_filter(nft, dt, tshift-dt)  # one sample gets lost in the shuffle

        # inverse fft, put in output array
        recT = ifft(recF)
        RF_out[c] = copy(recT.real[::-1])

    if normalize is not None:
        norm = 1 / np.max(np.abs(RF_out[normalize]))
        for RF in RF_out:
            RF *= norm

    return RF_out



# Copyright 2013-2022 Tom Eulenfeld, MIT license
"""
All objects and functions in rf  are after the Tom Eulenfeld's
rf project. License below.

The MIT License (MIT)

Copyright (c) 2013-2019 Tom Eulenfeld

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

"""
rf Documentation
================

rf is a Python framework for receiver function analysis.
Read and write support of necessary metadata is provided for
SAC, SeismicHandler and HDF5 waveform files.
Data is handled by the ``RFStream`` class which inherits a lot of useful
methods from its ObsPy ancestor ``Stream``,
but also has some unique methods necessary for receiver function calculation.


Method
------

The receiver function method is a popular technique to investigate crustal and
upper mantle velocity discontinuities. Basic concept of the method is that a
small part of incident P-waves from a teleseismic event gets converted to
S-waves at significant discontinuities under the receiver
(for P-receiver functions).
These converted Ps phases arrive at the station after the main P phase.
The response function of the receiver side (receiver function) is constructed
by removing the source and deep mantle propagation effects.
Firstly, the S-wave field is separated from the P-wave field by a rotation
from the station coordinate system (ZNE - vertical, north, east)
to the wave coordinate system (LQT - P-wave polarization,
approx. SV-wave polarization, SH-wave polarization). Alternatively, it is
possible to rotate to the ZRT coordinate system (vertical, radial, transverse).
Secondly, the waveform on the L component is deconvolved from the other
components, which removes source side and propagation effects.
The resulting functions are the Q and T component of the P receiver function.
Multiple reflected waves are also visible in the receiver function.
The conversion points of the rays are called piercing points.

For a more detailed description of the working flow see e.g. chapter 4.1 of
this_ dissertation.

.. image:: _static/2layer_rays.svg
   :height: 250px
   :alt: Ray paths

.. image:: _static/2layer_synrf.svg
   :height: 250px
   :alt: Synthetic receiver function

| *Left*: In a two-layer-model part of the incoming P-wave is converted to a
    S-wave at the layer boundary. Major multiples are Pppp, Ppps and Ppss.
| *Right*: Synthetic receiver function of Q component in a two-layer-model.

Installation
------------

Dependencies of rf are

    * ObsPy_ and some of its dependencies,
    * cartopy, geographiclib, shapely,
    * mtspec_ for multitaper deconvolution,
    * toeplitz_ for faster time domain deconvolution (optional),
    * obspyh5_ for hdf5 file support (optional),
    * tqdm for progress bar in batch processing (optional).

rf can be installed with ::

    pip install rf

The tests can be run with the script ::

    rf-runtests

To install the development version of rf download the source code and run ::

    pip install .

Here are some instructions to install rf into a fresh conda environment::

    conda config --add channels conda-forge
    conda create -n rfenv obspy cartopy geographiclib shapely h5py mtspec tqdm fortran-compiler
    conda activate rfenv
    pip install obspyh5 toeplitz rf
    rf-runtests

The ``fortran-compiler`` metapackage helps finding the right fortran compiler
and its dependencies for any OS. Often, problems with the compilation of Fortran
codes (toeplitz package) can be solved by setting
``export NPY_DISTUTILS_APPEND_FLAGS=1``.

Using the Python module
-----------------------

The main functionality is provided by the class `.RFStream`
which is derived from ObsPy's `~obspy.core.stream.Stream` class.

The canonical way to load a waveform file into a RFStream is to use
the `.read_rf()` function.

>>> from rf import read_rf
>>> stream = read_rf('myfile.SAC')

If you already have an ObsPy Stream and you want to turn it into a RFStream
use the generator of RFStream:

>>> from rf import RFStream
>>> stream = RFStream(obspy_stream)

The stream is again written to disc as usual by its write method:

>>> stream.write('outfile', 'SAC')

The RFStream object inherits a lot of useful methods from its ObsPy ancestor
(e.g. filter, taper, simulate, ...).

The module automatically maps important (for rf calculation) header information
from the stats object attached to every trace to the format specific headers.
At the moment only SAC and SH/Q headers are supported. When initializing an
RFStream the header information in the format specific headers are written to
the stats object and before writing the information stored in the stats object
is written back to the format specific headers. In this way the important
header information is guaranteed to be saved in the waveform files.
The following table reflects the mapping:

=================  =========  ======
stats              SH/Q       SAC
=================  =========  ======
station_latitude   COMMENT    stla
station_longitude  COMMENT    stlo
station_elevation  COMMENT    stel
event_latitude     LAT        evla
event_longitude    LON        evlo
event_depth        DEPTH      evdp
event_magnitude    MAGNITUDE  mag
event_time         ORIGIN     o
onset              P-ONSET    a
type               COMMENT    kuser0
phase              COMMENT    kuser1
moveout            COMMENT    kuser2
distance           DISTANCE   gcarc
back_azimuth       AZIMUTH    baz
inclination        INCI       user0
slowness           SLOWNESS   user1
pp_latitude        COMMENT    user2
pp_longitude       COMMENT    user3
pp_depth           COMMENT    user4
box_pos            COMMENT    user5
box_length         COMMENT    user6
=================  =========  ======

.. note::
    Q-file header COMMENT is used for storing some information, because
    the Q format has a shortage of predefined headers.

.. note::
    Alternatively the hdf5 file format can be used. It is supported via the
    obspyh5 package. In this case all supported stats
    entries are automatically attached to the stored data. The plugin also
    saves entries not listed here (e.g. processing information).

The first task when calculating receiver functions is calculating some ray
specific values like azimuth and epicentral distance. An appropriate stats
dictionary can be calculated with `.rfstats()`:

>>> from rf import rfstats
>>> stats = rfstats(station=station, event=event, phase='P', dist_range=(30,90))
>>> for tr in stream:
>>>     tr.stats.update(stats)

or if the station and event information is already stored in the stats object:

>>> rfstats(stream)

A typical workflow for P receiver function calculation and writing looks like

>>> stream.filter('bandpass', freqmin=0.05, freqmax=1.)
>>> stream.rf()
>>> stream.write('rf', 'Q')

rf can also calculate S receiver functions:

>>> stream.rf(method='S')

When calling stream.rf the following operations are performed depending on
the given kwargs:

    * filtering
    * trimming data to window relative to onset
    * downsampling
    * rotation
    * deconvolution

Please see `.RFStream.rf()`
for a more detailed description.
RFStream provides the possibility to perform moveout correction,
piercing point calculation and profile stacking.

Tutorials
---------

The following Jupyter notebooks can be viewed online or downloaded
to be locally reproduced.

    1. Calculate receiver functions - minimal example (notebook1_)
    2. Calculate receiver functions and stack them by common conversion points
       to create a profile (notebook2_)
    3. Calculate and compare receiver functions calculated with different
       deconvolution methods (notebook3_)

Command line tool for batch processing
--------------------------------------

The rf package provides a command line utility 'rf' which runs all the
necessary steps to perform receiver function calculation.
Use
``rf -h`` and ``rf {your_command} -h`` to discover the interface.
All you need is an inventory file (StationXML) and a file with events
(QuakeML) you want to analyze.

The command ::

    rf create

creates a :ref:`template configuration file <config_label>` in the current
directory. This file is in JSON format and well documented.
After adapting the file to your needs you can use the various
commands of rf to perform different tasks (e.g. receiver function
calculation, plotting).

To create the tutorial with a small included dataset and working configuration
you can use ::

    rf create --tutorial

Now start using rf ..., e.g. ::

    rf data calc myrf
    rf moveout myrf myrfmout
    rf plot myrfmout myrfplot
    rf --moveout-phase Psss moveout myrf myrfPsssmout

.. note::
    Development of the batch module has a lower priority than the Python API.

Miscellaneous
-------------

Please feel free to request features, report bugs or contribute code on
GitHub_. The code is continuously tested by Github Actions.

Citation
--------

If you found this package useful, please consider citing it.

Tom Eulenfeld (2020), rf: Receiver function calculation in seismology, *Journal of Open Source Software*, 5(48), 1808, doi: `10.21105/joss.01808`__.

.. __: https://doi.org/10.21105/joss.01808

.. _this: https://doi.org/10.17169/refubium-14424
.. _ObsPy: http://www.obspy.org/
.. _pip: http://www.pip-installer.org/
.. _obspyh5: https://github.com/trichter/obspyh5/
.. _toeplitz: https://github.com/trichter/toeplitz/
.. _mtspec: https://github.com/krischer/mtspec

.. _notebook1: http://nbviewer.jupyter.org/github/trichter/notebooks/blob/master/receiver_function_minimal_example.ipynb
.. _notebook2: http://nbviewer.jupyter.org/github/trichter/notebooks/blob/master/receiver_function_profile_chile.ipynb
.. _notebook3: https://nbviewer.jupyter.org/github/hfmark/notebooks/blob/main/rf_comparison.ipynb

.. _GitHub: https://github.com/trichter/rf/
"""
__version__ = '1.0.3-dev'

from rfsed.rf.profile import get_profile_boxes
from rfsed.rf.rfstream import read_rf, RFStream, rfstats
from rfsed.rf.util import iter_event_data, IterMultipleComponents

if 'dev' not in __version__:  # get image for correct version from travis-ci
    __doc__ = __doc__.replace('branch=master', 'branch=v%s' % __version__)



import numpy as np
import obspy
from obspy.io.sac import SACTrace
from obspy.signal.util import next_pow_2
from math import pi
from numpy.fft import fft, ifft, ifftshift
import matplotlib.pyplot as plt
from obspy import Trace, Stream
from scipy.interpolate import interp1d
from os.path import join, dirname, exists
from scipy.io import loadmat
from matplotlib.colors import ListedColormap
class ValueError(Exception): ...

def gaussFilter(dt, nft, gaussian):
    df = 1.0 / (nft * dt)
    nft21 = 0.5 * nft + 1
    f = df * np.arange(0, nft21)
    w = 2 * pi * f

    gauss = np.zeros([nft, 1])
    gauss1 = np.exp(-0.25 * (w / gaussian) ** 2) / dt
    gauss1.shape = (len(gauss1), 1)
    gauss[0:int(nft21)] = gauss1
    gauss[int(nft21):] = np.flipud(gauss[1:int(nft21) - 1])
    gauss = gauss[:, 0]

    return gauss

def gfilter(x, nfft, gauss, dt):
    Xf = fft(x, nfft)
    Xf = Xf * gauss * dt
    xnew = ifft(Xf, nfft).real
    return xnew

def correl(R, W, nfft):
    x = ifft(fft(R, nfft) * np.conj(fft(W, nfft)), nfft)
    x = x.real
    return x

def phaseshift(x, nfft, dt, tshift):
    Xf = fft(x, nfft)
    shift_i = int(tshift / dt)
    p = 2 * pi * np.arange(1, nfft + 1) * shift_i / nfft
    Xf = Xf * np.vectorize(complex)(np.cos(p), -np.sin(p))
    x = ifft(Xf, nfft) / np.cos(2 * pi * shift_i / nfft)
    x = x.real
    return x

def deconit(uin, win, dt, nt=None, tshift=10, gaussian=2.0, itmax=400, minderr=0.001, phase='P'):
    """
    Created on Wed Sep 10 14:21:38 2014

    In:
    uin = numerator (radial for PdS)
    win = denominator (vertical component for PdS)
    dt = sample interval (s)
    nt = number of samples
    tshift = Time until beginning of receiver function (s)
    gaussian = width of gaussian filter
    itmax = max # iterations
    minderr = Min change in error required for stopping iterations

    Out:
    RFI = receiver function
    rms = Root mean square error for predicting numerator after each iteration

    @author: Mijian Xu @ NJU
    """
    # print('Iterative Decon (Ligorria & Ammon):\n')
    if len(uin) != len(win):
        raise ValueError('The two input trace must be in same length')
    elif nt is None:
        nt = len(uin)
    else:
        pass

    rms = np.zeros(itmax)
    nfft = next_pow_2(nt)
    p0 = np.zeros(nfft)

    u0 = np.zeros(nfft)
    w0 = np.zeros(nfft)

    u0[0:nt] = uin
    w0[0:nt] = win

    gaussF = gaussFilter(dt, nfft, gaussian)
    # gaussF = _gauss_filter(dt, nfft, gaussian)

    u_flt = gfilter(u0, nfft, gaussF, dt)
    w_flt = gfilter(w0, nfft, gaussF, dt)

    wf = fft(w0, nfft)
    r_flt = u_flt

    powerU = np.sum(u_flt ** 2)

    it = 0
    sumsq_i = 1
    d_error = 100 * powerU + minderr
    maxlag = 0.5 * nfft
    # print('\tMax Spike Display is ' + str((maxlag) * dt))

    while np.abs(d_error) > minderr and it < itmax:
        rw = correl(r_flt, w_flt, nfft)
        rw = rw / np.sum(w_flt ** 2)

        if phase == 'P':
            i1 = np.argmax(np.abs(rw[0:int(maxlag) - 1]))
        else:
            i1 = np.argmax(np.abs(rw))
        amp = rw[i1] / dt

        p0[i1] = p0[i1] + amp
        p_flt = gfilter(p0, nfft, gaussF, dt)
        p_flt = gfilter(p_flt, nfft, wf, dt)

        r_flt = u_flt - p_flt
        sumsq = np.sum(r_flt ** 2) / powerU
        rms[it] = sumsq
        d_error = 100 * (sumsq_i - sumsq)

        sumsq_i = sumsq

        it = it + 1

    p_flt = gfilter(p0, nfft, gaussF, dt)
    p_flt = phaseshift(p_flt, nfft, dt, tshift)
    RFI = p_flt[0:nt]
    rms = rms[0:it - 1]

    return RFI, rms, it 
    

def deconwater(uin, win, dt, tshift=10., wlevel=0.05, gaussian=2.0, normalize=False, phase='P'):
    """
    Frequency-domain deconvolution using waterlevel method.

    :param uin: R or Q component for the response function
    :type uin: np.ndarray
    :param win: Z or L component for the source function
    :type win: np.ndarray
    :param dt: sample interval in second 
    :type dt: float
    :param tshift: Time shift before P arrival, defaults to 10.
    :type tshift: float, optional
    :param wlevel: Waterlevel to stabilize the deconvolution, defaults to 0.05
    :type wlevel: float, optional
    :param gaussian: Gauss factor, defaults to 2.0
    :type gaussian: float, optional
    :param normalize: If normalize the amplitude of the RF, defaults to False
    :type normalize: bool, optional

    :return: (rf, rms) RF and final rms.
    :rtype: (np.ndarray, float)
    """
    if uin.size != win.size:
        raise ValueError('The length of the \'uin\' must be same as the \'win\'')
    nt = uin.size
    nft = next_pow_2(nt)
    nfpts = nft / 2 + 1     # number of freq samples
    fny = 1. / (2.* dt);     # nyquist
    delf = fny / (0.5 * nft)
    freq = delf * np.arange(nfpts)
    w = 2 * pi * freq

    # containers
    # rff = np.zeros(nft); # Rfn in freq domain
    upf = np.zeros(nft); # predicted numer in freq domain

    # Convert seismograms to freq domain
    uf = fft(uin, nft)
    wf = fft(win, nft)

    # denominator
    df = wf * wf.conjugate()
    dmax = max(df.real);   

    # add water level correction
    phi1 = wlevel * dmax # water level
    # nwl = length( find(df<phi1) ) # number corrected
    df[np.where(df.real < phi1)[0]] = phi1
    gaussF = gaussFilter(dt, nft, gaussian)
    nf = gaussF * uf * wf.conjugate()

    # compute RF
    rff = nf / df
    
    # compute predicted numerator
    upf = rff * wf

    # add phase shift to RF
    w = np.append(w, -np.flipud(w[1:-1]))
    rff = rff * np.exp(-1j * w * tshift)

    # back to time domain
    rft = ifft(rff , nft)
    rft = rft[0:nt].real

    # compute the fit
    uf = gaussF * uf # compare to filtered numerator
    ut = ifft(uf, nft).real
    upt = ifft(upf, nft).real

    powerU = np.sum(ut[0:nt] ** 2)
    rms = np.sum((upt[0:nt] - ut[0:nt]) ** 2 )/powerU

    if normalize:
        gnorm = np.sum(gaussF) * delf * dt
        rft = rft.real / gnorm

    return rft, rms

def deconvolute(uin, win, dt, method='iter', **kwargs):
    if method.lower() == 'iter':
        return deconit(uin, win, dt, **kwargs)
    elif method.lower() == 'water':
        return deconwater(uin, win, dt, **kwargs)
    else:
        raise ValueError('method must be \'iter\' or \'water\'')

class RFTrace(obspy.Trace):
    def __init__(self, data=..., header=None):
        super().__init__(data=data, header=header)

    @classmethod
    def deconvolute(cls, utr, wtr, method='iter', **kwargs):
        header = utr.stats.__getstate__()
        for key, value in kwargs.items():
            header[key] = value
        if method.lower() == 'iter':
            rf, rms, it = deconit(utr.data, wtr.data, utr.stats.delta, **kwargs)
            header['rms'] = rms
            header['iter'] = it
        elif method.lower() == 'water':
            rf, rms = deconwater(utr.data, wtr.data, utr.stats.delta, **kwargs)
            header['rms'] = rms
            header['iter'] = np.nan
        else:
            raise ValueError('method must be \'iter\' or \'water\'')
        return cls(rf, header)
    
    
ei = 0+1j

def e_inverse(omega, rho, alpha, beta, p):
    """ E_inverse (Aki & Richards, pp. 161, Eq. (5.71))

    Parameters
    ----------
    omega : _type_
        _description_
    rho : _type_
        _description_
    alpha : _type_
        _description_
    beta : _type_
        _description_
    p : _type_
        _description_
    """
    e_inv = np.zeros([4,4], dtype=complex)
    eta = np.sqrt(1.0/(beta*beta) - p*p)
    xi  = np.sqrt(1.0/(alpha*alpha) - p*p)
    bp = 1.0 - 2.0*beta*beta*p*p

    e_inv[0,0] = beta*beta*p/alpha
    e_inv[0,1] = bp/(2.0*alpha*xi)
    e_inv[0,2] = -p/(2.0*omega*rho*alpha*xi) * ei
    e_inv[0,3] = -1.0/(2.0*omega*rho*alpha) * ei
    e_inv[1,0] = bp / (2.0*beta*eta)
    e_inv[1,1] = -beta*p
    e_inv[1,2] = -1.0/(2.0*omega*rho*beta) * ei
    e_inv[1,3] = p/(2.0*omega*rho*beta*eta) * ei
    e_inv[2,0] = e_inv[0,0]
    e_inv[2,1] = - e_inv[0,1]
    e_inv[2,2] = - e_inv[0,2]
    e_inv[2,3] = e_inv[0,3]
    e_inv[3,0] = e_inv[1,0]
    e_inv[3,1] = - e_inv[1,1]
    e_inv[3,2] = - e_inv[1,2]
    e_inv[3,3] = e_inv[1,3]
    return e_inv


def propagator_sol(omega, rho, alpha, beta, p, z):
    """ propagator (Aki & Richards, pp. 398, Eq. (3) in Box 9.1)

    Parameters
    ----------
    omega : _type_
        _description_
    rho : _type_
        _description_
    alpha : _type_
        _description_
    beta : _type_
        _description_
    p : _type_
        _description_
    z : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    
    p_mat = np.zeros([4,4], dtype=complex)
    beta2 = beta*beta
    p2 = p*p
    bp = 1.0 -2.0*beta2*p2
    eta = np.sqrt(1.0/(beta2) - p2)
    xi  = np.sqrt(1.0/(alpha*alpha) - p2)
    cos_xi = np.cos(omega*xi*z)
    cos_eta = np.cos(omega*eta*z)
    sin_xi = np.sin(omega*xi*z)
    sin_eta = np.sin(omega*eta*z)

    p_mat[0,0] = 2.0*beta2*p2*cos_xi + bp*cos_eta
    p_mat[0,1] = p*( bp/xi*sin_xi - 2.0*beta2*eta*sin_eta ) * ei
    p_mat[0,2] = (p2/xi*sin_xi + eta*sin_eta)/(omega*rho)
    p_mat[0,3] = p*(-cos_xi + cos_eta)/(omega*rho) * ei  
    p_mat[1,0] = p*( 2.0*beta2*xi*sin_xi - bp/eta*sin_eta ) * ei
    p_mat[1,1] = bp*cos_xi + 2.0*beta2*p2*cos_eta
    p_mat[1,2] = p_mat[0,3]
    p_mat[1,3] = (xi*sin_xi + p2/eta*sin_eta)/(omega*rho)
    p_mat[2,0] = omega*rho*( -4.0*beta2*beta2*p2*xi*sin_xi - bp*bp/eta*sin_eta )
    p_mat[2,1] = 2.0*omega*beta2*rho*p*bp*( cos_xi - cos_eta ) * ei
    p_mat[2,2] = p_mat[0,0]
    p_mat[2,3] = p_mat[1,0]
    p_mat[3,0] = p_mat[2,1]
    p_mat[3,1] = -omega*rho*( bp*bp/xi*sin_xi + 4.0*beta2*beta2*p2*eta*sin_eta  )
    p_mat[3,2] = p_mat[0,1]  
    p_mat[3,3] = p_mat[1,1]

    return p_mat

def haskell(omega, p, nl, ipha, alpha, beta, rho, h):
    i0 = 0
    e_inv = e_inverse(omega, rho[-1], alpha[-1], beta[-1], p)
    p_mat = propagator_sol(omega, rho[i0], alpha[i0], beta[i0], p, h[i0] )
    for i in range(i0+1, nl):
        p_mat2 = propagator_sol(omega, rho[i], alpha[i], beta[i], p, h[i])
        p_mat = np.matmul(p_mat2, p_mat)
    if nl > i0+1:
        sl = np.matmul(e_inv, p_mat)
    else:
        sl = e_inv
    denom = sl[2,0] * sl[3,1] - sl[2,1] * sl[3,0]
    if ipha >= 0:
        ur = sl[3,1] / denom
        uz = - sl[3,0] / denom
    else:
        ur = - sl[2,1] / denom
        uz = sl[2,0] / denom
    return ur, uz


def fwd_seis(rayp, dt, npts, ipha, alpha, beta, rho, h):
    nlay = h.size
    npts_max = next_pow_2(npts)
    ur_freq = np.zeros(npts_max, dtype=complex)
    uz_freq = np.zeros(npts_max, dtype=complex)
    nhalf = int(npts_max / 2 + 1)
    for i in range(1, nhalf):
        omg = 2*np.pi * i / (npts_max * dt)
        ur_freq[i], uz_freq[i] = haskell(omg, rayp, nlay, ipha, 
                                         alpha, beta, rho, h)
    ur = ifft(ur_freq).real[::-1]/npts_max
    uz = -ifft(uz_freq).real[::-1]/npts_max

    return ur[0:npts], uz[0:npts]


class SynRF():
    def __init__(self, depth, vp, vs, rho, rayp, dt, npts=2500, ipha=1, filter=None) -> None:
        """_summary_

        Parameters
        ----------
        depth : _type_
            np.array of depth
        vp : _type_
            np.array of vp
        vs : _type_
            np.array of vs
        rho : _type_
            np.array of rho
        rayp : _type_
            Ray-parameter in s/km
        dt : _type_
            Time interval
        npts : _type_
            samples of synthetic waveform
        ipha : _type_
            Specify incident wave 1 for P and -1 for S
        """
        self.depth = depth
        self.vp = vp
        self.vs = vs
        self.rho = rho
        self.dt = dt
        self.npts = npts
        self.thickness=[]
        for i in range(len(self.depth)):
            if i == 0:
                thick=self.depth[i]
                self.thickness.append(thick)
            else:
                if i==1:
                    thick=self.depth[i]-self.depth[i-1]
                    self.thickness.append(thick)
                else:
                    if i==2:
                        thick=self.depth[i]-self.depth[i-1]
                        self.thickness.append(thick)
                    else:
                        if i==3:
                            thick=self.depth[i]-self.depth[i-1]
                            self.thickness.append(thick)
                        else:
                            if i==4:
                                thick=self.depth[i]-self.depth[i-1]
                                self.thickness.append(thick)
                            else:
                                if i==5:
                                    thick=self.depth[i]-self.depth[i-1]
                                    self.thickness.append(thick)
        self.thickness=np.array(self.thickness)
        if not isinstance(rayp, (float, list, np.ndarray)):
            raise TypeError('The rayp should be in float, list and np.ndarray')
        if isinstance(rayp, float):
            self.rayp = [rayp]
        else:
            self.rayp = rayp
        self.ipha = ipha

    def run_fwd(self):
        """Forward modelling synthetic seismograms.

        ``SynSeis.rstream`` and ``SynSeis.zstream`` are generated as 
        radial and vertical Seismograms in ``Obspy.Stream`` type.
        """
        self.rstream = Stream()
        self.zstream = Stream()
        for _, rayp in enumerate(self.rayp):
            ur, uz = fwd_seis(rayp, self.dt, self.npts, self.ipha,
                            self.vp, self.vs, self.rho,
                            self.thickness)
            tr = Trace(data=ur)
            tr.stats.delta = self.dt
            self.rstream.append(tr)
            tr = Trace(data=uz)
            tr.stats.delta = self.dt
            self.zstream.append(tr)

    def filter(self, freqmin, freqmax, order=2, zerophase=True):
        """Apply a bandpass filter on synthetic waveforms

        Parameters
        ----------
        freqmin : float
            Minimum cut-off frequency
        freqmax : float
            maximum cut-off frequency
        order : int, optional
            Order of filter, by default 2
        zerophase : bool, optional
            whether use a zero-phase filter, by default True
        """
        for st in [self.rstream, self.zstream]:
            st.filter('bandpass', freqmin=freqmin, freqmax=freqmax,
                      corners=order, zerophase=zerophase)

    def run_deconvolution(self, pre_filt=[0.05, 2], shift=10, gaussian=2.0, **kwargs):
        if pre_filt is not None:
            self.filter(*pre_filt)
        rfstream = Stream()
        for i, _ in enumerate(self.rayp):
            rftr = RFTrace.deconvolute(self.rstream[i], self.zstream[i], tshift=shift,
                                       gaussian=gaussian, **kwargs)
            rfstream.append(rftr)
        return rfstream


import unittest
import warnings
from rfsed import hkZhu
import numpy as np
from rfsed.util import rfMoho_example, rfSed_example, save_plot, save_tests
try:
    import matplotlib
except ImportError:
    Matplotlib = None

class TesthkZhu(unittest.TestCase):
    def test_hk(self):
        rfstream = rfMoho_example()
        staname='OPLO'
        rfstreams = rfstream.select(component='R', station=staname)
        K= np.linspace(1.65,1.95,121)
        H=np.linspace(20,60,201)
        Vp=6.9
        w1, w2, w3 =[0.6, 0.3, 0.1]
        self.assertTrue(hkZhu.hk(rfstreams, layer='Moho', stack=True, w1 = w1, w2 = w2, w3 = w3, K= K, H=H, Vp = Vp))
    
    def test_plothk(self):
        rfstream = rfMoho_example()
        staname='OPLO'
        rfstreams = rfstream.select(component='R', station=staname)
        K= np.linspace(1.65,1.95,121)
        H=np.linspace(20,60,201)
        Vp=6.9
        w1, w2, w3 =[0.6, 0.3, 0.1]
        savepath=save_tests()
        HKResult= hkZhu.hk(rfstreams, layer='Moho', stack=True, w1 = w1, w2 = w2, w3 = w3, K= K, H=H, Vp = Vp)
        hkZhu.plothk(HKResult=HKResult, g = [75.,10., 15., 2.5], rmneg = None,savepath=savepath, format = 'jpg')

def suite():
    return unittest.makeSuite(TesthkZhu, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')


# Copyright 2013-2019 Tom Eulenfeld, MIT license
"""
All objects and functions below are after the Tom Eulenfeld's
rf project License below.

The MIT License (MIT)

Copyright (c) 2013-2019 Tom Eulenfeld

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
"""
Simple move out and piercing point calculation.
"""
from math import floor
from pkg_resources import resource_filename

import numpy as np
from rfsed.rf.util import direct_geodetic, DEG2KM


_MODEL_CACHE = {}


def load_model(fname='iasp91'):
    """
    Load model from file.

    :param fname: path to model file or 'iasp91'
    :return: `SimpleModel` instance

    The model file should have 4 columns with depth, vp, vs, n.
    The model file for iasp91 starts like this::

        #IASP91 velocity model
        #depth  vp    vs   n
         0.00  5.800 3.360 0
         0.00  5.800 3.360 0
        10.00  5.800 3.360 4
    """
    try:
        return _MODEL_CACHE[fname]
    except KeyError:
        pass
    fname_key = fname
    if fname == 'iasp91':
        fname = resource_filename('rfsed', 'rf/data/iasp91.dat')
    values = np.loadtxt(fname, unpack=True)
    try:
        z, vp, vs, n = values
        n = n.astype(int)
    except ValueError:
        n = None
        z, vp, vs = values
    _MODEL_CACHE[fname_key] = model = SimpleModel(z, vp, vs, n)
    return model


def _interpolate_n(val, n):
    vals = [np.linspace(val[i], val[i + 1], n[i + 1] + 1, endpoint=False)
            for i in range(len(val) - 1)]
    return np.hstack(vals + [np.array([val[-1]])])


class SimpleModel(object):

    """
    Simple 1D velocity model for move out and piercing point calculation.

    :param z: depths in km
    :param vp: P wave velocities at provided depths in km/s
    :param vs: S wave velocities at provided depths in km/s
    :param n: number of support points between provided depths

    All arguments can be of type numpy.ndarray or list.
    """

    def __init__(self, z, vp, vs, n=None):
        assert len(z) == len(vp) == len(vs)
        if n is not None:
            z = _interpolate_n(z, n)
            vp = _interpolate_n(vp, n)
            vs = _interpolate_n(vs, n)
        self.z = z[:-1]
        self.dz = np.diff(z)
        self.vp = vp[:-1]
        self.vs = vs[:-1]
        self.t_ref = {}

    def calculate_vertical_slowness(self, slowness, phase='PS'):
        """
        Calculate vertical slowness of P and S wave.

        :param slowness: slowness in s/deg
        :param phase: Weather to calculate only P, only S or both vertical
            slownesses
        :return: vertical slowness of P wave, vertical slowness of S wave
            at different depths (z attribute of model instance)
        """
        phase = phase.upper()
        hslow = slowness / DEG2KM  # convert to horizontal slowness (s/km)
        qp, qs = 0, 0
        # catch warnings because of negative root
        # these values will be nan
        with np.errstate(invalid='ignore'):
            if 'P' in phase:
                qp = np.sqrt(self.vp ** (-2) - hslow ** 2)
            if 'S' in phase:
                qs = np.sqrt(self.vs ** (-2) - hslow ** 2)
        return qp, qs

    def calculate_delay_times(self, slowness, phase='PS'):
        """
        Calculate delay times between direct wave and converted phase.

        :param slowness: ray parameter in s/deg
        :param phase: Converted phase or multiple (e.g. Ps, Pppp)
        :return: delay times at different depths
        """
        phase = phase.upper()
        qp, qs = self.calculate_vertical_slowness(slowness, phase=phase)
        dt = (qp * phase.count('P') + qs * phase.count('S') -
              2 * (qp if phase[0] == 'P' else qs)) * self.dz
        return np.cumsum(dt)

    def stretch_delay_times(self, slowness, phase='Ps', ref=6.4):
        """
        Stretch delay times of provided slowness to reference slowness.

        First, calculate delay times (time between the direct wave and
        the converted phase or multiples at different depths) for the provided
        slowness and reference slowness.
        Secondly, stretch the the delay times of provided slowness to reference
        slowness.

        :param slowness: ray parameter in s/deg
        :param phase: 'Ps', 'Sp' or multiples
        :param ref: reference ray parameter in s/deg
        :return: original delay times, delay times stretched to reference
            slowness
        """
        if len(phase) % 2 == 1:
            msg = 'Length of phase (%s) should be divisible by two'
            raise ValueError(msg % phase)
        phase = phase.upper()
        try:
            t_ref = self.t_ref[phase]
        except KeyError:
            self.t_ref[phase] = t_ref = self.calculate_delay_times(ref, phase)
        t = self.calculate_delay_times(slowness, phase)
        if phase[0] == 'S':
            t_ref = -t_ref
            t = -t
        try:
            index = np.nonzero(np.isnan(t))[0][0] - 1
        except IndexError:
            index = len(t)
        t = t[:index]
        t_ref = t_ref[:index]
        return (np.hstack((-t[1:10][::-1], t)),
                np.hstack((-t_ref[1:10][::-1], t_ref)))

    def moveout(self, stream, phase='Ps', ref=6.4):
        """
        In-place moveout correction to reference slowness.

        :param stream: stream with stats attributes onset and slowness.
        :param phase: 'Ps', 'Sp', 'Ppss' or other multiples
        :param ref: reference slowness (ray parameter) in s/deg
        """
        for tr in stream:
            st = tr.stats
            if not (st.starttime <= st.onset <= st.endtime):
                msg = 'onset time is not between starttime and endtime of data'
                raise ValueError(msg)
            index0 = int(floor((st.onset - st.starttime) * st.sampling_rate))
            t0, t1 = self.stretch_delay_times(st.slowness, phase=phase,
                                              ref=ref)
            S_multiple = phase[0].upper() == 'S' and len(phase) > 3
            if S_multiple:
                time0 = st.starttime - st.onset + index0 * st.delta
                old_data = tr.data[:index0][::-1]
                t = -time0 - np.arange(index0) * st.delta
                new_t = -np.interp(-t, -t0, -t1, left=0, right=0)
                data = np.interp(-t, -new_t, old_data, left=0., right=0.)
                tr.data[:index0] = data[::-1]
            else:
                if t0[-1] > t1[-1]:
                    index0 += 1
                time0 = st.starttime - st.onset + index0 * st.delta
                old_data = tr.data[index0:]
                t = time0 + np.arange(len(tr) - index0) * st.delta
                # stretch old times to new times
                new_t = np.interp(t, t0, t1, left=0, right=None)
                # interpolate data at new times to data samples
                data = np.interp(t, new_t, old_data, left=None, right=0.)
                tr.data[index0:] = data
        return stream

    def ppoint_distance(self, depth, slowness, phase='S'):
        """
        Calculate horizontal distance between piercing point and station.

        :param depth: depth of interface in km
        :param slowness: ray parameter in s/deg
        :param phase: 'P' or 'S' for P wave or S wave. Multiples are possible.
        :return: horizontal distance in km
        """
        if len(phase) % 2 == 0:
            msg = 'Length of phase (%s) should be even'
            raise ValueError(msg % phase)
        phase = phase.upper()
        xp, xs = 0., 0.
        qp, qs = self.calculate_vertical_slowness(slowness, phase=phase)
        if 'P' in phase:
            xp = np.cumsum(self.dz * slowness / DEG2KM / qp)
        if 'S' in phase:
            xs = np.cumsum(self.dz * slowness / DEG2KM / qs)
        x = xp * phase.count('P') + xs * phase.count('S')
        z = self.z
        index = np.nonzero(depth < z)[0][0] - 1
        return x[index] + ((x[index + 1] - x[index]) *
                           (depth - z[index]) / (z[index + 1] - z[index]))

    def ppoint(self, stats, depth, phase='S'):
        """
        Calculate latitude and longitude of piercing point.

        Piercing point coordinates and depth are saved in the pp_latitude,
        pp_longitude and pp_depth entries of the stats object or dictionary.

        :param stats: Stats object or dictionary with entries
            slowness, back_azimuth, station_latitude and station_longitude
        :param depth: depth of interface in km
        :param phase: 'P' for piercing point of P wave, 'S' for piercing
            point of S wave. Multiples are possible, too.
        :return: latitude and longitude of piercing point
        """
        dr = self.ppoint_distance(depth, stats['slowness'], phase=phase)
        lat = stats['station_latitude']
        lon = stats['station_longitude']
        az = stats['back_azimuth']
        plat, plon = direct_geodetic((lat, lon), az, dr)
        stats['pp_depth'] = depth
        stats['pp_latitude'] = plat
        stats['pp_longitude'] = plon
        return plat, plon



import unittest
import warnings
from rfsed import hkSeqYeck
from rfsed.hkSeqYeck import hkSeq, plotSeqhk
import numpy as np
from rfsed.util import rfMoho_example, rfSed_example, save_plot, save_tests
try:
    import matplotlib
except ImportError:
    Matplotlib = None

class hkSeqYeck(unittest.TestCase):
    def test_hkSeq(self):
        rfstSed = rfSed_example()
        staname='OPLO'
        rfstreamSed = rfstSed.select(component='R', station=staname)
        KSed= np.linspace(1.65,2.25,201)
        HSed=np.linspace(0,10,201)
        VpSed=2.5
        w1Sed, w2Sed, w3Sed = [0.6, 0.3, 0.1]
        rfstMoho = rfMoho_example()
        rfstreamMoho = rfstMoho.select(component='R', station=staname)
        KMoho= np.linspace(1.65,1.95,121)
        HMoho=np.linspace(20,60,201)
        VpMoho=6.9
        w1Moho, w2Moho, w3Moho = [0.6, 0.3, 0.1]
        self.assertTrue(hkSeq(rfstreamSed, rfstreamMoho,  w1Sed = w1Sed, 
             w2Sed = w2Sed, w3Sed=w3Sed, KSed=KSed, HSed=HSed, VpSed=VpSed,
             w1Moho = w1Moho, w2Moho = w2Moho, w3Moho=w3Moho, KMoho=KMoho, 
             HMoho=HMoho, VpMoho=VpMoho, stack = False))
    def test_plotSeqhk(self):
        rfstSed = rfSed_example()
        staname='OPLO'
        rfstreamSed = rfstSed.select(component='R', station=staname)
        KSed= np.linspace(1.65,2.25,201)
        HSed=np.linspace(0,10,201)
        VpSed=2.5
        w1Sed, w2Sed, w3Sed = [0.6, 0.3, 0.1]
        rfstMoho = rfMoho_example()
        rfstreamMoho = rfstMoho.select(component='R', station=staname)
        KMoho= np.linspace(1.65,1.95,121)
        HMoho=np.linspace(20,60,201)
        VpMoho=6.9
        w1Moho, w2Moho, w3Moho = [0.6, 0.3, 0.1]
        savepath=save_tests()
        SequentialHKResult=hkSeq(rfstreamSed, rfstreamMoho,  w1Sed = w1Sed, 
             w2Sed = w2Sed, w3Sed=w3Sed, KSed=KSed, HSed=HSed, VpSed=VpSed,
             w1Moho = w1Moho, w2Moho = w2Moho, w3Moho=w3Moho, KMoho=KMoho, 
             HMoho=HMoho, VpMoho=VpMoho, stack = False)
        plotSeqhk(SequentialHKResult=SequentialHKResult, savepath=savepath)
        

def suite():
    return unittest.makeSuite(hkSeqYeck, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')


# Copyright 2013-2019 Tom Eulenfeld, MIT license
"""
All objects and functions below are after the Tom Eulenfeld's
rf project License below.

The MIT License (MIT)

Copyright (c) 2013-2019 Tom Eulenfeld

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
"""
Functions for receiver function plotting.
"""
import warnings

import matplotlib.patheffects as PathEffects
from matplotlib.ticker import (AutoMinorLocator, FixedLocator, FixedFormatter,
                               MaxNLocator)
import matplotlib.pyplot as plt
import numpy as np


def _label(stream):
    label_fmts = ['{network}.{station}.{location}.{channel}',
                  '{network}.{station}.{location}.{channel:.2}?',
                  '{network}.{station}']
    for label_fmt in label_fmts:
        labelset = {label_fmt.format(**tr.stats) for tr in stream}
        if len(labelset) == 1:
            return labelset.pop()
    return ''


def plot_rf(stream, fname=None, fig_width=7., trace_height=0.5,
            stack_height=0.5, dpi=None,
            scale=1, fillcolors=(None, None), trim=None,
            info=(('back_azimuth', u'baz (°)', 'C0'),
                  ('distance', u'dist (°)', 'C3')),
            show_vlines=False):
    """
    Plot receiver functions.

    :param stream: stream to plot
    :param fname: filename to save plot to. Can be None. In this case
        the figure is left open.
    :param fig_width: width of figure in inches
    :param trace_height: height of one trace in inches
    :param stack_height: height of stack axes in inches
    :param dpi: dots per inch for the created figure
    :param scale: scale for individual traces
    :param fillcolors: fill colors for positive and negative wiggles
    :param trim: trim stream relative to onset before plotting using
         `~.rfstream.RFStream.slice2()`
    :param info: Plot one additional axes showing maximal two entries of
        the stats object. Each entry in this list is a list consisting of
        three entries: key, label and color.
        info can be None. In this case no additional axes is plotted.
    :param show_vlines: If True, show vertical alignment grid lines on plot
        at positions of the major x-tick marks.
    """

    if len(stream) == 0:
        return
    if trim:
        stream = stream.slice2(*trim, reftime='onset')
    if info is None:
        info = ()
    N = len(stream)
    # calculate axes and figure dimensions
    # big letters: inches, small letters: figure fraction
    H = trace_height
    HS = stack_height
    FB = 0.5
    FT = 0.2
    DW = 0.1
    FH = H * (N + 2) + HS + FB + FT + DW
    h = H / FH
    hs = HS / FH
    fb = FB / FH
    ft = FT / FH
    FL = 0.5
    FR = 0.2
    FW = fig_width
    FW3 = 0.8
    FW2 = FW - FL - FR - (DW + FW3) * bool(info)
    fl = FL / FW
    fr = FR / FW
    fw2 = FW2 / FW
    fw3 = FW3 / FW
    # init figure and axes
    fig = plt.figure(figsize=(FW, FH), dpi=dpi)
    ax1 = fig.add_axes([fl, fb, fw2, h * (N + 2)])
    if info:
        ax3 = fig.add_axes(
            [1 - fr - fw3, fb, fw3, h * (N + 2)], sharey=ax1)
        info = list(info)
        info[0] = [ax3] + list(info[0])
        if len(info) > 1:
            ax4 = ax3.twiny()
            info[1] = [ax4] + list(info[1])
    # plot individual receiver functions

    def _plot(ax, t, d, i):
        c1, c2 = fillcolors
        if c1:
            ax.fill_between(t, d + i, i, where=d >= 0, lw=0., facecolor=c1)
        if c2:
            ax.fill_between(t, d + i, i, where=d < 0, lw=0., facecolor=c2)
        ax.plot(t, d + i, 'k')
    xlim = (0, 0)
    max_ = max(np.max(np.abs(tr.data)) for tr in stream)
    for i, tr in enumerate(stream):
        times = tr.times(reftime=tr.stats.onset)
        xlim = (min(xlim[0], times[0]), max(xlim[1], times[-1]))
        _plot(ax1, times, tr.data / max_ * scale, i + 1)
    # plot right axes with header information
    for ax, header, label, color in info:
        data = [tr.stats[header] for tr in stream]
        ax.plot(data, 1 + np.arange(len(stream)), '.' + color, mec=color)
        ax.set_xlabel(label, color=color, size='small')
        if header == 'back_azimuth':
            ax.set_xticks(np.arange(5) * 90)
            ax.set_xticklabels(['0', '', '180', '', '360'], size='small')
        else:
            ax.xaxis.set_major_locator(MaxNLocator(4))
            for l in ax.get_xticklabels():
                l.set_fontsize('small')
        ax.xaxis.set_minor_locator(AutoMinorLocator())
    # set x and y limits
    ax1.set_xlim(*xlim)
    ax1.set_ylim(-0.5, N + 1.5)
    ax1.set_yticklabels('')
    ax1.set_xlabel('time (s)')
    ax1.xaxis.set_minor_locator(AutoMinorLocator())
    aligner_color = "#a0a0a080"
    if show_vlines:
        ax1.xaxis.grid(True, color=aligner_color, linestyle=':')

    # plot stack
    try:
        stack = stream.stack()
    except ValueError:
        msg = 'Different npts for traces in one RF plot. Do not plot stack.'
        warnings.warn(msg)
    else:
        if len(stack) > 1:
            warnings.warn('Different stations or channels in one RF plot. ' +
                          'Do not plot stack.')
        elif len(stack) == 1:
            ax2 = fig.add_axes([fl, 1 - ft - hs, fw2, hs], sharex=ax1)
            _plot(ax2, times, stack[0].data, 0)
            for l in ax2.get_xticklabels():
                l.set_visible(False)
            ax2.yaxis.set_major_locator(MaxNLocator(4))
            for l in ax2.get_yticklabels():
                l.set_fontsize('small')
            if show_vlines:
                ax2.xaxis.grid(True, color=aligner_color, linestyle=':')
    # annotate plot with seed id
    bbox = dict(boxstyle='round', facecolor='white', alpha=0.8, lw=0)
    title = '%s traces  %s' % (len(stream), _label(stream))
    ax1.annotate(title, (1 - 0.5 * fr, 1 - 0.5 * ft),
                 xycoords='figure fraction', va='top', ha='right',
                 bbox=bbox, clip_on=False)
    # save plot
    if fname:
        fig.savefig(fname, dpi=dpi)
        plt.close(fig)
    else:
        return fig


def _get_geoaxes(crs=None, latlons=None):
    """Return cartopy geoaxis"""
    if crs is None:
        from cartopy.crs import AzimuthalEquidistant
        latlon0 = np.median(latlons, axis=0)
        crs = AzimuthalEquidistant(*latlon0[::-1])
    return plt.axes(projection=crs)


def __pc():
    from cartopy.crs import PlateCarree as PC
    return PC()


def plot_stations(inventory, label_stations=True, ax=None, crs=None, **kwargs):
    """
    Plot stations.

    :param inventory: station inventory
    :param label_stations: weather to label stations
    :param ax: geoaxes (default None: new ax will be created)
    :param crs: coordinate reference system for new geoaxis, (default: None,
        then AzimuthalEquidistant projection with appropriate center is used.)
    :param \*\*kwargs: other kwargs are passed to ax.scatter() call
    """
    try:
        # assume inventory to be stream
        data = [((tr.stats.station_latitude, tr.stats.station_longitude),
                 tr.stats.station) for tr in inventory]
        latlons, names = zip(*list(set(data)))
    except AttributeError:
        # inventory is Inventory
        latlons, names = zip(*[((sta.latitude, sta.longitude), sta.code)
                               for net in inventory for sta in net])
    if ax is None:
        ax = _get_geoaxes(crs=crs, latlons=latlons)
    kw = dict(s=200, marker='v', c='darkred', linewidth=0.5, zorder=3)
    kw.update(kwargs)
    ax.scatter(*list(zip(*latlons))[::-1], transform=__pc(), **kw)
    if label_stations:
        path_effect = PathEffects.withStroke(linewidth=3, foreground="white")
        kw = {'xycoords': __pc()._as_mpl_transform(ax),
              'xytext': (10, 0), 'textcoords': 'offset points', 'zorder': 4,
              'path_effects': [path_effect]}
        for latlon, name in zip(latlons, names):
            ax.annotate(name, latlon[::-1], **kw)
    return ax


def plot_ppoints(ppoints, inventory=None, label_stations=True, ax=None,
                 crs=None, **kwargs):
    """
    Plot piercing points with stations.

    :param ppoints: list of (lat, lon) tuples of piercing points
    :param inventory, label_stations: plot stations, see `plot_stations`
    :param ax: geoaxes (default None: new ax will be created)
    :param crs: coordinate reference system for new geoaxis, (default: None,
        then AzimuthalEquidistant projection with appropriate center is used.)
    :param \*\*kwargs: other kwargs are passed to ax.scatter() call
    """
    if ax is None:
        ax = _get_geoaxes(crs=crs, latlons=ppoints)
    if inventory is not None:
        plot_stations(inventory, label_stations=label_stations, ax=ax)
    kw = dict(s=50, marker='x', color='k', alpha=0.2, zorder=2)
    kw.update(kwargs)
    ax.scatter(*list(zip(*ppoints))[::-1], transform=__pc(), **kw)
    return ax


def plot_profile_map(boxes, inventory=None, label_stations=True, ppoints=None,
                     ax=None, crs=None, **kwargs):
    """
    Plot profile map with stations and piercing points.

    :param boxes: boxes created with `~.profile.get_profile_boxes()`
    :param inventory, label_stations: plot stations, see `plot_stations`
    :param ppoints: list of (lat, lon) tuples of piercing points,
        see `plot_ppoints`
    :param ax: geoaxes (default None: new ax will be created)
    :param crs: coordinate reference system for new geoaxis, (default: None,
        then AzimuthalEquidistant projection with appropriate center is used.)
    :param \*\*kwargs: other kwargs are passed to ax.add_geometries() call
    """
    if ax is None:
        latlons = [boxes[len(boxes)//2]['latlon']]
        ax = _get_geoaxes(crs=crs, latlons=latlons)
    if inventory is not None:
        plot_stations(inventory, label_stations=label_stations, ax=ax)
    if ppoints is not None:
        plot_ppoints(ppoints, ax=ax)
    kw = dict(facecolor='none', edgecolor='0.8', zorder=1)
    kw.update(kwargs)
    for box in boxes:
        ax.add_geometries([box['poly']], crs=__pc(), **kw)
    return ax


def plot_profile(profile, fname=None, figsize=None, dpi=None,
                 scale=1, fillcolors=('C3', 'C0'),
                 trim=None, top=None, moveout_model='iasp91'):
    """
    Plot receiver function profile.

    :param profile: stream holding the profile
    :param fname: filename to save plot to. Can be None. In this case
        the figure is left open.
    :param figsize: figsize of the created figure
    :param dpi: dots per inch for the created figure
    :param scale: scale for individual traces
    :param fillcolors: fill colors for positive and negative wiggles
    :param trim: trim stream relative to onset before plotting using
         `~.rfstream.RFStream.slice2()`
    :param top: show second axes on top of profile with additional information.
        Valid values: 'hist' - Plot histogram showing the number of receiver
        functions stacked in the corresponding bin
    :param moveout_model: string with model filename. Will be loaded into a
        `~.simple_model.SimpleModel` object to calculate depths for
        tick labels.
    """
    if len(profile) == 0:
        return
    if trim:
        profile = profile.slice2(*trim, reftime='onset')
    fig = plt.figure(figsize=figsize, dpi=dpi)
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.7])
    widths = [tr.stats.box_length for tr in profile]
    pad = max(1, scale) * min(widths)
    xlim = (min(tr.stats.box_pos for tr in profile) - pad,
            max(tr.stats.box_pos for tr in profile) + pad)
    max_ = max(np.max(np.abs(tr.data)) for tr in profile)
    for tr in profile:
        x = tr.stats.box_pos + scale * tr.data / max_ * min(widths)
        y = tr.times() - (tr.stats.onset - tr.stats.starttime)
        ax.plot(x, y, 'k')
        c1, c2 = fillcolors
        if c1:
            ax.fill_betweenx(y, x, tr.stats.box_pos,
                             where=x >= tr.stats.box_pos, facecolor=c1)
        if c2:
            ax.fill_betweenx(y, x, tr.stats.box_pos,
                             where=x < tr.stats.box_pos, facecolor=c2)
    ax.set_xlabel('distance (km)')
    ax.set_ylim(max(y), min(y))
    ax.set_ylabel('time (s)')
    if moveout_model:
        from rf.simple_model import load_model
        model = load_model(moveout_model)
        phase = profile[0].stats.moveout
        slowness = profile[0].stats.slowness
        pd = model.calculate_delay_times(phase=phase, slowness=slowness)
        ax2 = ax.twinx()
        ax.get_shared_y_axes().join(ax, ax2)
        dkm = 50
        if profile[0].stats.endtime - profile[0].stats.onset > 50:
            dkm = 200
        d1 = np.arange(20) * dkm
        d2 = np.arange(100) * dkm / 5
        t1 = np.interp(d1, model.z, pd)
        t2 = np.interp(d2, model.z, pd)
        myLocator = FixedLocator(t1)
        myMinorLocator = FixedLocator(t2)
        myFormatter = FixedFormatter([str(i) for i in d1])
        ax2.yaxis.set_major_locator(myLocator)
        ax2.yaxis.set_minor_locator(myMinorLocator)
        ax2.yaxis.set_major_formatter(myFormatter)
        ax2.set_ylabel('depth (km)')
        ax2.set_ylim(ax.get_ylim())
    if top is not None:
        ax3 = fig.add_axes([0.1, 0.85, 0.8, 0.1], sharex=ax)
    if top == 'hist':
        left = [tr.stats.box_pos - tr.stats.box_length / 2 for tr in profile]
        height = [tr.stats.num for tr in profile]
        ax3.bar(left, height, widths, color='cadetblue')
        plt.setp(ax3.get_xticklabels(), visible=False)
        ax3.spines['top'].set_color('none')
        ax3.spines['right'].set_color('none')
        ax3.spines['left'].set_color('none')
        ax3.xaxis.set_ticks_position('bottom')
        ax3.yaxis.set_ticks_position('left')
        ax3.set_yticks(ax3.get_ylim())
    elif top is not None:
        raise NotImplementedError("'%s' not supported for top parameter" % top)
    ax.set_xlim(*xlim)
    if fname:
        fig.savefig(fname, dpi=dpi)
        plt.close(fig)
    else:
        return fig
